
#include <math.h>
#include <stdlib.h>
#include "iml_port.h"
#include "gml_gfit.h"

static const IML_DOUBLE E_CONST = 2.71828182846;
static const IML_DOUBLE ROOT2   = 1.41421356237;
//=============================================================================
inline IML_DOUBLE weibull(IML_DOUBLE C, IML_DOUBLE A, IML_DOUBLE B, IML_DOUBLE Base=E_CONST) {
	return(1.0 - 0.5*pow(Base, -1 * pow( C/A, B) ) );
	}

inline IML_DOUBLE weibull_b2(IML_DOUBLE C, IML_DOUBLE A, IML_DOUBLE B) {
	return(1.0 - 0.5*pow(2, -1 * pow( C/A, B) ) );
	}


//===============================================================================
// CUMULATIVE NORMAL - based on complementary erf in Numerical Recipes in C p 176
//===============================================================================
inline double cnorm(IML_DOUBLE X, IML_DOUBLE Mu, IML_DOUBLE SDev) {
 double T,Z,Ans;

	// MAKE SURE THAT WE USE THE ROOT-2 FOR TRUE CNORM
	X = (X-Mu)/(SDev*ROOT2);
	Z = fabs(X);
	T = 1.0/(1.0+0.5*Z);
	Ans = T*exp( -Z*Z - 1.26551223 +
			T*(1.00002368 + T*(0.37409196 + T*(0.09678418 +
			T*(-0.18628806 + T*(0.27886807 + T*(-1.13520398 +
			T*(1.48851587 + T*(-0.82215223 + T*0.17087277)))))))));

	Ans = (X>=0.0)?(1.0-Ans):1.0-(2.0-Ans);
	// NOW 'ans' IS IN THE RANGE -1 .. 1
	// WHEN WE RETURN IT, WE CONVERT TO THE RANGE [0..1] SO THAT IT IS
	// ACTUALLY THE CUMULATIVE NORMAL
	return(0.5+0.5*Ans);

} // end cnorm

inline IML_DOUBLE DB_to_C(IML_DOUBLE DB) { return(pow(10,-fabs(DB)/20.0)); };

inline IML_DOUBLE C_to_DB(IML_DOUBLE C) { return(20*log10(C)); };

inline double uniform(float Min=0, float Max=1) { 	
	srand(rand()); srand(rand()); 
	return( Min + (Max-Min)*rand()/(double)RAND_MAX); 
};


//=============================================================================
// FIRST WE DEFINE A DATA STRUCTURE WHICH WILL HOLD THE DATA AND 
// INTERMEDIATE MODEL VALUES.
//=============================================================================
class my_data {

	public:
		float ContrastDB; 			// INDEPENDENT VARIABLE
		float Contrast; 			// INDEPENDENT VARIABLE ALIAS
		
		float TrialCount;			// THESE TWO TOGETHER PRODUCE A SINGLE
		float NumberCorrect; 		// (MEASURED) DEPENDENT VARIABLE

		float NumberInCorrect; 		// THESE TWO ARE JUST ALIASES FOR THE 
		float DataFractionCorrect;	// DEPENDENT VARIABLE
		
		float ModelFractionCorrect;	// THE MODEL ESTIMATE FOR THE DEPENDENT VARIABLE
		
		float ConfidenceWeight;		// THIS ALSO COMES FROM MEASUREMENT (OR ESTIMATE)
		
		float BestLogLikelihood;	// INTERMEDIATE VAR; WE ARE MINIMIZING CHI-SQUARE
		
		float PriorC;
		float PriorI;

		void add_trial(IML_LINT CorrectFlag) {
			TrialCount++;
			if(CorrectFlag == TRUE) NumberCorrect++; else NumberInCorrect++;
			DataFractionCorrect = NumberCorrect/TrialCount;
			set_best_LL();
			}

		// THIS IS THE FUNCTION UPON WHICH GRADIENT DESCENT (OR SOMETHING SIMILAR) 
		// WILL BE PERFORMED. THIS WILL ONLY BE CALLED BY THE evaluate() FUNCTION WHICH
		// MUST BE DEFINED BY THE CORRESPONDING fitter
		IML_DOUBLE logLikelihood(IML_FLOAT PiorMag=1)	
			{ return(	-(NumberCorrect+PriorC*PiorMag)*log(ModelFractionCorrect)
						 -(NumberInCorrect+PriorI*PiorMag)*log(1.0-ModelFractionCorrect) );  };  
		
		void set_best_LL(void) { 
			if(TrialCount && NumberCorrect && NumberInCorrect)
				BestLogLikelihood =	-NumberCorrect*log(DataFractionCorrect) 
									- NumberInCorrect*log(1.0-DataFractionCorrect ); 
			else BestLogLikelihood = 0;				
			};  
		
};

//=============================================================================
//=============================================================================
class my_fitter : public gml_generic_fitter {
	public:

		my_data IML_PTR Data;
		
		IML_LINT TotalTrials;
		
		IML_DOUBLE	PriorWeight;
		IML_DOUBLE	PriorBeta;
		IML_DOUBLE	PriorAlpha;
		IML_DOUBLE	PriorMagnitude;
		
		IML_LINT 	StartDB;
		IML_LINT		DeltaDB;
		
		IML_DOUBLE BestLogLikelihood;
		
		// WE KNOW THAT THERE WILL ONLY BE TWO PARAMETERS
		my_fitter(IML_USINT MaxDataCnt) : gml_generic_fitter(2) {

			MaxDataCnt = (MaxDataCnt > 0 ? MaxDataCnt : 1);
			DataCount = MaxDataCnt;
			Data = new my_data[MaxDataCnt];
			BestLogLikelihood = 0;
			
			PriorBeta = 3;
			PriorAlpha = 20;
			PriorMagnitude = 1;
			TotalTrials = 0;
			}; // end

	
		virtual IML_DOUBLE model_error(register IML_ULINT I) 	
			{  return(Data[I].ModelFractionCorrect - Data[I].DataFractionCorrect); };
		virtual IML_DOUBLE data(register IML_ULINT I) 	
			{  return(Data[I].DataFractionCorrect); };
		virtual IML_DOUBLE weight(register IML_ULINT I) 
			{ return(Data[I].ConfidenceWeight); };

			
		virtual IML_DOUBLE evaluate( register gml_double_vec &PVector) { 
		
			Evaluations++;
		
			for(register IML_LINT Index=0; Index < DataCount; Index++) {
				my_data &DP = Data[Index];
				DP.ModelFractionCorrect = weibull(DP.Contrast, PVector[0], PVector[1]);				
				DP.ModelFractionCorrect = GML_PFLOOR(0.001,DP.ModelFractionCorrect);
				DP.ModelFractionCorrect = GML_PCEIL(0.999,DP.ModelFractionCorrect);
				} // end for
				
			// HERE WE CALCULATE THE LOGLIKELIHOOD *INCLUDING THE PRIOR*
			register IML_DOUBLE LL=0;
			for(register IML_LINT Index=0; Index < DataCount; Index++)
				LL += Data[Index].logLikelihood(PriorWeight);
				
			// AND WE RETURN THE (PUTATIVE) CHI-SQUARE VALUE
			return( 2*(LL-BestLogLikelihood) ); 
			}; // end evaluate

		void initialize(IML_LINT StDB, IML_LINT DelDB, 
				IML_FLOAT PriorAlpha, IML_FLOAT PriorBeta, IML_FLOAT PriorN) {
			StartDB = StDB;
			DeltaDB = DelDB;
			for(IML_LINT I=0; I<DataCount; I++) {
				my_data &DP = Data[I];
				DP.ContrastDB = StartDB+I*DeltaDB; // contrast in 2db steps
				DP.Contrast = DB_to_C(DP.ContrastDB); 
				DP.TrialCount=0;				
				DP.NumberCorrect=0; 		
				DP.NumberInCorrect=0; 
				DP.ConfidenceWeight = 1.0;
				DP.PriorC = PriorN*weibull_b2(DP.Contrast, DB_to_C(PriorAlpha), PriorBeta );
				DP.PriorI = PriorN-DP.PriorC;
				} // end for
			}; // END initialize

		void add_trial(IML_LINT BinDB, IML_LINT CorrectFlag) {
			IML_ULINT Index = (BinDB-StartDB)/DeltaDB;
			if(Index > DataCount) Index = DataCount-1;
			Data[Index].add_trial(CorrectFlag);
			TotalTrials++;
			PriorWeight = pow(2, -1 * pow( TotalTrials/PriorAlpha, PriorBeta) );
			} // end add trial
			
		IML_LINT next_guess(void) {
			fitdata();
			IML_FLOAT NewC = parameter_value(0);
			NewC = C_to_DB(NewC)-0.0001; // get decibels; account for numerical roundoff
			IML_ULINT Index = long((NewC-StartDB)/DeltaDB + 0.5);
			if(Index > DataCount) Index = DataCount-1;
			return(Index);
			};


		IML_DOUBLE bestLL(void) { 
			BestLogLikelihood=0;
			for(register IML_LINT Index=0; Index < DataCount; Index++) {
				Data[Index].set_best_LL();
				BestLogLikelihood += Data[Index].BestLogLikelihood;
				} // END FOR
			return(BestLogLikelihood); 
			}; // end bestLL

};


//===============================================================================
// main program:  example of how to use the generic fitter
//===============================================================================
int main(void) {

	IML_LINT DataCount = 20;
	
	// ALLOCATE A FITTER FOR MAX OF 10 PIECES OF DATA
	my_fitter DataFitter(DataCount);
	
	// INITIALIZE BIN VALUES; KEEP THE DEFAULT PRIOR DECAY STUFF
	
	DataFitter.initialize(
		-45, 	// start bin value (db)
		2, 	// delta bin value (db)
		-20, 	// prior threshold
		2, 	// prior slope
		10	// give a weight to the prior
		);
	
	// THE SIMULATION PSYCHOMETRIC FUNCTION PARAMETERS
	float RealBeta = 3;					// real slope
	float RealAlpha = DB_to_C(-25);	// real threshold 
	srand(-1);
	// simulation trials
	IML_LINT TrialCnt = 100;
	// FAKE SOME GUESSES

	// SET CONVERGENCE CRITETIA:  ANSWER TOLERANCE AND MAX-POWELL ITERATIONS
	DataFitter.set_convergence(1e-4, 15);
	
	// CONFIGURE THE PARAMETERS FOR START
	IML_FLOAT CurrentC = DB_to_C(-20);
							//  ID		 NAME			MIN    	MAX   GUESS		OPTIMIZE-FLAG
	DataFitter.configure_parameter(0, "Weibull Alpha",	0.001,	0.99,	CurrentC,	TRUE);
	DataFitter.configure_parameter(1, "Weibull Beta",	0.1,	10.0,		2,		FALSE);

while(TrialCnt--) {

	// fake a trial
	IML_FLOAT Pc = weibull_b2(CurrentC, RealAlpha, RealBeta);
	IML_FLOAT Test = uniform();
	IML_LINT CorrectFlag = Pc > Test; // generate user response
	
	IML_LINT DB_2 = C_to_DB(CurrentC)-0.001;
	DataFitter.add_trial( DB_2, CorrectFlag);
	
	// OPTIMUM VALUE IS ALSO CHI-SQUARE BECAUSE THATS OUR OBJECTIVE FUNCTION;
	IML_LINT Index = DataFitter.next_guess();
	
	// add random +- 2,4 dB (+/- 1,2 bins)
	IML_LINT Select = long(uniform(0.0,4.9999)); //{0,1,2,3,4}
	Index += Select - 2;
	if(Index < 0) Index = 0;
	if(Index >= DataCount) Index = DataCount-1;
	CurrentC = DataFitter.Data[Index].Contrast; // convert co contrast.
	}

/*
	fprintf(stderr,"Evaluations =%d; LineMinimizations = %d; PowellIterations = %d;\n",
						DataFitter.Evaluations, 
						DataFitter.LineMinimizations, 
						DataFitter.PowellIterations
						);
*/

	// SET CONVERGENCE CRITETIA:  ANSWER TOLERANCE AND MAX-POWELL ITERATIONS
	DataFitter.set_convergence(1e-10, 30);

	// CALCULATE THIS FIRST SO evaluate() RETURNS CHI-SQUARE
	IML_DOUBLE BestLogLikelihood = DataFitter.bestLL();
	DataFitter.optimize_flag(1) = TRUE;
	IML_DOUBLE OptimumValue = 	DataFitter.fitdata();

	IML_LINT DOF = 0;
	for(int I=0; I<DataCount; I++) {
		if(DataFitter.Data[I].TrialCount) {
			fprintf(stderr,"DB %f;  T %f;  NC %f;\n",
					DataFitter.Data[I].ContrastDB,
					DataFitter.Data[I].TrialCount,
					DataFitter.Data[I].NumberCorrect);
			DOF++;
			}
		}

	// PRINT THE CHISQUARE FIT
	fprintf(stderr,"Chisquare fit of %f with %d DOF\n",OptimumValue, DOF-2);
	
	// GRAB THE OPTIMUM VALUES
	IML_DOUBLE BestAlpha = DataFitter.parameter_value(0);
	IML_DOUBLE BestBeta = DataFitter.parameter_value(1);
	fprintf(stderr,"OPT: a,b %f,%f\n",BestAlpha,BestBeta);
	
	DataFitter.calculate_covariance_information();
	
	DataFitter.calculate_fit_statistics();
	
	// CALCULATE THE RIGOROUS 95% CHI-SQUARE LIMITS
	IML_DOUBLE Limit;  IML_LINT MaxSteps = 15;
	IML_DOUBLE Contour95 = OptimumValue + 3.84;
	Limit = DataFitter.find_parameter_CI( 
			gml_generic_fitter::CI_HI, // WHICH BOUNDARY TO FIND
			Contour95, 						// WHAT CONTOUR LEVEL TO FIND
			0, 								// PARAMETER ID
			BestAlpha, 							// LOWER BRACKET GUESS
			0.99, 							// UPPER BRACKET GUESS
			0.01, 							// ANSWER TOLLERANCE
			MaxSteps);						// MAX ITERATIONS TO TAKE
		
	fprintf(stderr,"alpha high is %f\n",Limit);

	Limit = DataFitter.find_parameter_CI(
		gml_generic_fitter::CI_LO, Contour95, 0, 0.001, BestAlpha, 0.01, MaxSteps);
	fprintf(stderr,"alpha low is %f\n",Limit);

	Limit = DataFitter.find_parameter_CI(
		gml_generic_fitter::CI_HI, Contour95, 1, BestBeta, 10, 0.01, MaxSteps);
	fprintf(stderr,"beta high is %f\n",Limit);

	Limit = DataFitter.find_parameter_CI(
		gml_generic_fitter::CI_LO, Contour95, 1, .1, BestBeta, 0.01, MaxSteps);
	fprintf(stderr,"beta low is %f\n",Limit);


return(0);
}
