#ifndef __IML_GFIT_H__
#define __IML_GFIT_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/


#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "gml_nrcv.h"


#define SQUARE_IT(x) ((Temp = (x))*Temp)
#define GML_PFLOOR(x,n) ((x)<(n)?(n):(x))
#define GML_PCEIL(x,n) ((x)>(n)?(n):(x))

typedef const char *  ptr_to_const_char;

/**  Base class for model fitting.   This class provides the interface and basic tools
for fitting linear or nonlinear models to data.  If you can write the code for
the model, then you can probably use this.  
<p> The current fitter implements Powell's
method as it is outlined in NRC.  The error function is to be defined in a derived
(child) class along with a few other things.  This base class provides other tools
in addition to a method for finding the optimal parameter values (Powell's method),
namely, goodness of fit measures, and two methods for
determining error bounds on the optimal parameter estimates.</p>
*/
class gml_generic_fitter {

	protected:

		// THESE ARE FOR FINDING THE SECOND PARTIALS OF THE ERROR SURFACE
		// WITH THE INTENT OF FINDING THE COVARIANCE MATRIX (THE INVERSE OF THE HESSIAN)
		// ugly global parameters resulting from NRC routines
		gml_double_vec SamplePoint;
		gml_double_vec TestParms;
		
		IML_DOUBLE H_Fraction;

		//iml_float_mat *Hessian;
		//iml_float_mat *Covariance;
		//iml_float_mat *Covariance_LU_DComp;
		//iml_float_mat *Correlation;
		//iml_float_mat *Error;
		//iml_sint_vec  *ParameterIDs;

		float **Hessian;
		float **Covariance;
		float **Covariance_LU_DComp;
		float **Correlation;
		float **Error;
		int  *ParameterIDs;

		// GLOBAL VARIABLES TO CONTROL 'Powell()' BEHAVIOR
		IML_LINT MaxIters;
		IML_DOUBLE Epsilon;

		//	 GLOBAL VARIABLE FOR INDICATING THE PARAMETERS OVER WHICH WE WILL
		//  BE OPTIMIZING FOR ANY GIVEN SESSION.
		gml_lint_vec OptimizeFlags;

		ptr_to_const_char * ParmName;
		gml_double_vec MinParmV;
		gml_double_vec MaxParmV;
		gml_double_vec OptimizedParms;		// STORAGE FOR COMPUTED OPTIMAL PARAMETERS
		IML_DOUBLE ObjectiveFunctionMinimum;

		void scale_parms(IML_DOUBLE F, gml_double_vec &Parm);
		void check_limits(gml_double_vec &Parm);


		IML_DOUBLE line_minimization(
			gml_double_vec &XStart, gml_lint_vec &Subspace, gml_double_vec &Direction);
		void powell(gml_double_vec &XStart, gml_lint_vec &Subspace);

		int find_contour( gml_double_vec &Parms, IML_LINT ParmID, IML_DOUBLE a, IML_DOUBLE b, IML_DOUBLE k, IML_DOUBLE eps);
		unsigned short count_fitted_parms(gml_lint_vec &Subspace );

	public:

		/**@name A:  What to do in your derived (child) class.
		<p> There are several actions to take when producing a specific fitter
		from this base class.  This includes properly calling the constructor
		for this base class, setting the value of a specific
		variable and over-riding several class methods (writing the contents of
		specific functions) that are needed by the tools provided in this
		base class. </p>
		*/
		//@{

		/**@name Constructor to call.*/
		//@{
		/** The base class constructor.  No other constructors are provided, and there
		is no default value for the parameter count.
		@param PCount The model parameter count.
		@return None (constructors don't return values).
		*/
		gml_generic_fitter(IML_USINT PCount);
		//@}

		/**@name Parameter to set.*/
		//@{
		/** Data count in the fitting process.  This value must be set so that
		Degrees of Freedom can be calculated correctly.
		*/
		IML_ULINT DataCount;
		// Maximum number of data values (used for error checking).
		//IML_ULINT MaxCount;
		//@}

		/**@name Methods to over-ride for <em>Powell's method</em>.*/
		//@{
		/** Evaluate the model error given specific model parameters.
		This virtual method is called by the implementation of <em>Powell's method</em>
		("method" used differently in these two senses)
		and should calculate some metric of the error between the model and
		the data.   This could be the sum-of-squared-error for each data point,
		or perhaps the chi-squared value if we are doing a maximum likelihood
		fit.
		@param Parameters is a gml_double_vec reference filled with parameter
		values selecting a specific instance of the model family.
		@return The return value should be the "error" of the model relative to 
		the data.
		@see Numerical Recipes in C for explanation of the chi-square statistic and
		its relationship to  squared error.
		*/
		virtual IML_DOUBLE evaluate(register gml_double_vec &Parameters) { return(0); };
		//@}
		
		/**@name Methods to over-ride for fit statistics. */
		//@{
		/// For data point I, the difference between data and model.
		virtual IML_DOUBLE model_error(register IML_ULINT I) 	{ return(0); };
		/// The value of data point I (assuming scalar data).
		virtual IML_DOUBLE data(register IML_ULINT I) 	{ return(0); };
		/// The weight of data point I (should be defaulted to 1).
		virtual IML_DOUBLE weight(register IML_ULINT I) { return(0); };
		//@}
		
		//@}

		/**@name B:  Fitting, convergence and parameter configuration. */
		//@{
		/// Run the fitter on the data.
		IML_DOUBLE fitdata(void);
		
		/// Set the tolerance and maximum number of iterations for Powell's method.
		void set_convergence(IML_DOUBLE Eps, IML_LINT MaxI)
			{ Epsilon = Eps; MaxIters = MaxI; }

		/** Use this method to completely configure a single parameter.
		*/
		void configure_parameter(IML_USINT I, ptr_to_const_char Name, 
				IML_DOUBLE Min, IML_DOUBLE Max, IML_DOUBLE Value,
				IML_LINT OptimizeOrNot = TRUE) { 
			if(I >= ParmCount) I = ParmCount-1;
			ParmName[I]=""; 
			MinParmV[I] = Min; 
			MaxParmV[I] = Max; 
			OptimizedParms[I] = Value; 
			OptimizeFlags[I] = OptimizeOrNot;			
			};

		/// Configure or access parameter name.
		ptr_to_const_char &parameter_name(IML_USINT I) {
			if(I >= ParmCount) I = ParmCount-1;
			return(ParmName[I]);
			};
			
		/// Configure or access parameter maximum allowed value.
		IML_DOUBLE &parameter_max(IML_USINT I) {
			if(I >= ParmCount) I = ParmCount-1;
			return(MaxParmV[I]);
			};
			
		/// Configure or access parameter minimum allowed value.
		IML_DOUBLE &parameter_min(IML_USINT I) {
			if(I >= ParmCount) I = ParmCount-1;
			return(MinParmV[I]);
			};
			
		/// Configure or access current parameter value.
		IML_DOUBLE &parameter_value(IML_USINT I) {
			if(I >= ParmCount) I = ParmCount-1;
			return(OptimizedParms[I]);
			};
			
		/// Configure or access whether a parameter will be optimized.
		IML_LINT &optimize_flag(IML_USINT I) {
			if(I >= ParmCount) I = ParmCount-1;
			return(OptimizeFlags[I]);
			};

		// Configure the number of data points in the fit.
//		void set_data_count(IML_USINT Cnt) {
//			Cnt = (Cnt > MaxCount ? MaxCount : Cnt);
//			DataCount = Cnt;
//			DegreesOfFreedom = DataCount - ParmCount;
//			};
		//@}
			

		/**@name C:  Calculating standard fit statistics.*/
		//@{
		/// Once fitdata() is called, this calculales the standard goodness-of-fit statistics.
		void calculate_fit_statistics(void);
		//@}
		
		/**@name D:  Unweighted fit statistics.*/
		//@{
		
		///  See McKay (bla bla bla).
		IML_DOUBLE PseudoOccamFactor;

		/// Data mean.
		IML_FLOAT AlgebraicMean;
		/// Data standard deviation.
		IML_FLOAT StandardDeviation;
		/// The coefficient of determination for the model.
		IML_FLOAT CoeffOfDetermination;
		/// The r^2 for the model.
		IML_FLOAT RSquared;
		/// The Model Selection Criterion for the model.
		IML_FLOAT MSC;
		//@}
		
		/**@name E:  Weighted fit statistics.*/
		//@{
		/// Weighted data mean.
		IML_FLOAT WeightedMean;
		/// Weighted data standard deviation.
		IML_FLOAT WStandardDeviation;
		/// Weighted coefficient of determination for the model.
		IML_FLOAT WCoeffOfDetermination;
		/// Weighted r^2 for the model.
		IML_FLOAT WRSquared;
		/// Weighted Model Selection Criterion for the model.
		IML_FLOAT WMSC;
		//@}
		

		/**@name F:  Calculating standard error bounds on parameters. */
		//@{

		/** Calculate information from the evaluate() (error) surface
		for use in parameter error bounds.  First the Hessian of the evaluate() (error) 
		surface is calculated, together with a (lower bound) estimate on the amount of
		error for each entry in the matrix.  Then the covariance and correlation matrices 
		for the parameters is calculated from the Hessian.
		*/
		void calculate_covariance_information(void);
		
		/// Access the covariance value for two parameters.
		IML_FLOAT covariance(IML_USINT I, IML_USINT J) {
			if(!Covariance) return(0);
			if(I >= CovMatrixRank) I = CovMatrixRank-1;
			if(J >= CovMatrixRank) J = CovMatrixRank-1;
			return(Covariance[I+1][J+1]);
			};
			
		/// Access the correlation value for two parameters.
		IML_FLOAT correlation(IML_USINT I, IML_USINT J) {
			if(!Correlation) return(0);
			if(I >= CovMatrixRank) I = CovMatrixRank-1;
			if(J >= CovMatrixRank) J = CovMatrixRank-1;
			return(Correlation[I+1][J+1]);
			};
		
		/// Access the entries in the Hessian.
		IML_FLOAT hessian(IML_USINT I, IML_USINT J) {
			if(!Hessian) return(0);
			if(I >= CovMatrixRank) I = CovMatrixRank-1;
			if(J >= CovMatrixRank) J = CovMatrixRank-1;
			return(Hessian[I+1][J+1]);
			};
		
		/// Access the estimates of (lower bound) error for the Hessian entries.
		IML_FLOAT error(IML_USINT I, IML_USINT J) {
			if(!Error) return(0);
			if(I >= CovMatrixRank) I = CovMatrixRank-1;
			if(J >= CovMatrixRank) J = CovMatrixRank-1;
			return(Error[I+1][J+1]);
			};
		//@}

		/**@name G:  Calculating robust error bounds via error-surface search. */
		//@{
		// Enumerated type for use with find_parameter_CI().
		enum {CI_LO = -1, CI_HI = 1};
		/// Method for finding error bounds on parameters via evaluate() surface contours.
		double find_parameter_CI(IML_LINT UpOrDown, IML_DOUBLE Contour, IML_LINT TargetParm,
				IML_DOUBLE LowBracket, IML_DOUBLE HighBracket, IML_DOUBLE Tolerance, IML_LINT & MaxSteps);
		//@}

		/**@name H:  Informational variables. */
		//@{
		/// This parameter is set at construction time, and is what it says.
		IML_ULINT ParmCount;
		/// This parameter is set when <em>fitdata()</em> is called.
		IML_ULINT FittedParmCnt;
		/// This parameter is set when <em>calculate_covariance_information()</em> is called.
		IML_ULINT CovMatrixRank;
		/// This parameter is DataCount - FittedParmCount.
		IML_LINT DegreesOfFreedom;

		///  Hessian determinant for singular-matrix checking.
		IML_DOUBLE Determinant;

		/** The total number of calls to the model evaluation function.  This must
		be set in the evaluate() function call defined in the child class.
		*/
		IML_LINT Evaluations;
		/// Total number of sub-optimizations in all iterations of Powell's method.
		IML_LINT LineMinimizations;
		/// Total number of iterations of Powell's method to reach the minimum.
		IML_LINT PowellIterations;
		//@}


}; // end class gml_generic_fitter


#endif // __IML_GFIT_H__
