#ifndef __IML_NRCV_H__
#define __IML_NRCV_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include "iml_port.h"

/**  A template class for Numerical-Recipes-like vectors.
The class encapsulates the vector operations one might want, but is by
no means complete (yet).  I originally hoped to be able to use this to
replace NRC's utilities, but alas it was not to be.  On the other hand,
these are nicer because we have bounds checking and simplified notation.
Note that this class is crude compared to what can be found elsewhere
(i.e. I make absolutely no claims about elegance here) but it does the
job. :-)
*/
template <class VType> class gml_nrc_vector {
	public:
		VType *Vec; // MUST THIS STAY AS THE *FIRST* ELEMENT?
		
		IML_LINT Length;
		
		/// The only constructor.
		gml_nrc_vector(long Len)  { Vec = new VType[Len]; Length=Len; };
			
		// The desctructor just cleans up the array, if it was allocated.
		~gml_nrc_vector() { if(Vec) delete [] Vec; };
		
//		operator VType*(void) {return(Vec - 1);}
			
		VType & operator[](long I) { return(Vec[I]); };
		
		/// Safe (bounds checked) reading and writing of vector values.
		VType & operator()(long I) { if(I>Length)I=Length-1; if(I<0)I=0;return(Vec[I]); };
		
		/// Dot product operator returning a scalar.
		VType operator*(gml_nrc_vector<VType> & V) { // dot product
			register VType Acc=0;
			for(register IML_LINT I=0;I<Length; I++) Acc += Vec[I] * V[I]; 
			return(Acc);
			};
			
		/// Multiply the vector by a double, returning a reference to myself.
		gml_nrc_vector<VType> & operator*=(IML_DOUBLE F) 
			{ for(register IML_LINT I=0;I<Length; I++) Vec[I] *= F; return(*this); };
			
		/// Element-wise assignment from a vector, returning a reference to myself.
		gml_nrc_vector<VType> & operator=(gml_nrc_vector &V) 
			{ for(register IML_LINT I=0;I<Length; I++) Vec[I] = V.Vec[I]; return(*this); };
			
		/// Element-wise assignment of a constant, returning a reference to myself.
		gml_nrc_vector<VType> & operator=(IML_DOUBLE K) 
			{ for(register IML_LINT I=0;I<Length; I++) Vec[I] = K; return(*this); };
			
		/// Element-wise addition from a vector, returning a reference to myself.
		gml_nrc_vector<VType> & operator+=(gml_nrc_vector &V) 
			{ for(register IML_LINT I=0;I<Length; I++) Vec[I] += V.Vec[I]; return(*this); };
			
		/// Element-wise subtraction from a vector, returning a reference to myself.
		gml_nrc_vector<VType> & operator-=(gml_nrc_vector &V) 
			{ for(register IML_LINT I=0;I<Length; I++) Vec[I] -= V.Vec[I]; return(*this); };

}; // end class definition
	



/*  A template class for Numerical-Recipes-like 2-D matrices.
The class encapsulates the matrix operations one might want, but is by
no means complete (yet).  I originally hoped to be able to use this to
replace NRC's utilities, but alas it was not to be.  There is
*/
template <class MType> class gml_nrc_sqrmat {
	public:
		gml_nrc_vector<MType> **Mat;
		MType ** NRCMat;
		IML_LINT Length1, Length2;
		
		operator MType**(void) {return(NRCMat-1);}
		
		gml_nrc_vector<MType> &operator[](long I) { return( *Mat[I]); };
		
		gml_nrc_sqrmat(long Len1, long Len2) { 
			Length1=Len1; Length2=Len2;
			Mat = new gml_nrc_vector<MType>*[Len1];
			NRCMat = new MType*[Len1];
			for(long I=0; I<Len1; I++) {
				Mat[I] = new gml_nrc_vector<MType>(Len2);
				MType *Temp = (MType*)Mat[I];
				NRCMat[I] = Temp - 1;
				} // end for
			};
			
		~gml_nrc_sqrmat() { 
			if(Mat) {
				for(long I=0; I<Length1; I++) delete Mat[I];
				delete [] Mat; 
				} // end if
			if(NRCMat) delete [] NRCMat;
			};

		gml_nrc_sqrmat<MType> & operator*=(IML_DOUBLE F) 
			{ for(register IML_LINT I=0;I<Length1; I++) {
				gml_nrc_vector<MType> *Temp = Mat[I];
				Temp->operator*=(F); 
				} // end for
			return(*this);
			};
			
		gml_nrc_sqrmat<MType> &  operator=(IML_DOUBLE K) 
			{ for(register IML_LINT I=0;I<Length1; I++) Mat[I]->operator=(K); return(*this); };
			
		gml_nrc_sqrmat<MType> &  operator=(gml_nrc_sqrmat &M) 
			{ for(register IML_LINT I=0;I<Length1; I++) Mat[I]->operator=(M[I]); return(*this); };
			
		gml_nrc_sqrmat<MType> &  operator+=(gml_nrc_sqrmat &M) 
			{ for(register IML_LINT I=0;I<Length1; I++) Mat[I]->operator+=(M[I]); return(*this); };
			
		gml_nrc_sqrmat<MType> &  operator-=(gml_nrc_sqrmat &M) 
			{ for(register IML_LINT I=0;I<Length1; I++) Mat[I]->operator-=(M[I]); return(*this); };
			
}; // end class definition
	
/** A vector of doubles. 
@see gml_nrc_vector */
typedef gml_nrc_vector<IML_DOUBLE> gml_double_vec;
/** A vector of floats. 
@see gml_nrc_vector */
typedef gml_nrc_vector<IML_FLOAT> gml_float_vec;
/** A vector of signed 32-bit integers. 
@see gml_nrc_vector */
typedef gml_nrc_vector<IML_LINT> gml_lint_vec;
/** A vector of signed 16 bit integers. 
@see gml_nrc_vector */
typedef gml_nrc_vector<IML_SINT> gml_sint_vec;

typedef gml_nrc_sqrmat<IML_DOUBLE> gml_double_mat;
typedef gml_nrc_sqrmat<IML_FLOAT> gml_float_mat;
typedef gml_nrc_sqrmat<IML_LINT> gml_lint_mat;
typedef gml_nrc_sqrmat<IML_SINT> gml_sint_mat;

#endif  // __IML_NRCV_H__
