//============================================================================
class filter_2d {
	public:
		virtual void operator()(iml_fcomplex_image &FilterTarget) = 0;
	};


//============================================================================
//============================================================================
float drift_image(iml_ubyte_image &ByteImage, 
	float Speed, float Direction, filter_2d &Filter );


//============================================================================
class max_radial_filter : public filter_2d {
		float MaxRadialSquared;
	public:
	
		// max frequency is in cycles per pixel
		max_radial_filter(float Max) 
			: MaxRadialSquared(Max*Max) {};
			
		virtual void operator() (iml_fcomplex_image &FilterTarget);
	};

//============================================================================
//============================================================================
void max_radial_filter::operator() (iml_fcomplex_image &FilterTarget) {

	if(FilterTarget.dimensionality() != 2 ) 
		throw("max_radial_filter requires a 2d FilterTarget");

	long XSize 	= FilterTarget.hyperplane_dimension_size(0);
	long YSize	= FilterTarget.hyperplane_dimension_size(1);
	
	long FirstX = FilterTarget.hyperplane_dimension_start(0);
	long LastX  = FirstX + XSize - 1;
	long FirstY = FilterTarget.hyperplane_dimension_start(1);
	long LastY  = FirstY + YSize - 1;
	
	iml_fcomplex Zero(0.0, 0.0);

		// change magnitude and phase
	for(long FY = FirstY; FY<=LastY; FY++) {
		for(long FX = FirstX; FX<=LastX; FX++) {
			// convert x and y to cycles per pixel
			float X = FX/float(XSize), Y=FY/float(YSize);
			float RadiusSquared = X*X + Y*Y;
			if( RadiusSquared >= MaxRadialSquared)
				FilterTarget.hp_fa(FX,FY) = Zero;
			} // end for
		} // end for


}; // end max_radial_filter::operator()



//============================================================================
//
//============================================================================
float drift_image(iml_ubyte_image &ByteImage, 
	float Speed /* pixels/frame */, float Direction /* deg */, filter_2d &Filter ) {

	if(ByteImage.dimensionality() != 3 ) throw("drift_image requires a 3d byteimage input");

	long XSize 		= ByteImage.hyperplane_dimension_size(0);
	long YSize		= ByteImage.hyperplane_dimension_size(1);
	long SeqLength	= ByteImage.hyperplane_dimension_size(2);

	// xnought and ynought are in image fractions per frame
	// x and y are scaled separately
	float XNought = Speed/float(XSize) * cos(Direction * IML_RAD_PER_DEG);
	float YNought = Speed/float(YSize) * sin(Direction * IML_RAD_PER_DEG);
	

	iml_fcomplex_image Template(XSize,YSize);
	iml_fcomplex_image WorkSpace(XSize,YSize);
	if(  !Template.valid_memory() || !WorkSpace.valid_memory() ) {
		iml_error("drift_image","Could not allocate working space");
		return(-1);
		} // end if

	// move image from first frame to template
	for(long K=0; K<Template.element_count();K++) {
		Template[K] = iml_fcomplex(ByteImage[K],0);
		} //e nd for

	// move it to the fourier domain
	Template.forward_FFT(TRUE);
	
	// filter the template as requested
	Filter(Template);
	
	long FirstX = Template.hyperplane_dimension_start(0);
	long LastX  =  FirstX + XSize - 1;
	long FirstY = Template.hyperplane_dimension_start(1);
	long LastY  =  FirstY + YSize - 1;

	// re-code Template as magnitude and phase
	for(long K=0; K<Template.element_count();K++)
		Template[K] = iml_fcomplex( Template[K].get_mag(), Template[K].get_phase() );
	
	// prepare to work with one frame at a time
	ByteImage.set_hyperplane_size(2);
	
//MacVideo.display_image(ByteImage, mvl_screen_position(200-XSize-1,200));

	mvl_clock Timer;
	Timer.start();
	
	// make SeqLength frames
	for(long Frame = 0; Frame<SeqLength; Frame++) {
		
		// set to current frame
		ByteImage.select_hyperplane(Frame);
				
		// make sure workspace and template have same index ranges
		WorkSpace.fourier_domain();

		// change magnitude and phase
		for(long FY = FirstY; FY<=LastY; FY++)
			for(long FX = FirstX; FX<=LastX; FX++) {
			
				float NextPhase = Template.hp_fa(FX,FY).Imag +
							IML_2PI*Frame*(FX*XNought + FY*YNought);
				
				float Mag = Template(FX,FY).Real;
				float Real = Mag * cos(NextPhase);
				float Imag = Mag * sin(NextPhase);

				// turn back to real/imag
				WorkSpace.hp_fa(FX,FY) = iml_fcomplex(Real, Imag);
				} // end for
		
			// inverse FFT
			WorkSpace.inverse_FFT(TRUE);
		
			// assign to proper image scan line
			for(long K=0; K<WorkSpace.element_count();K++)  
				ByteImage.hp_raw_fa(K) = WorkSpace[K].Real;
//			if(Frame==(0))
//			MacVideo.display_image(ByteImage, mvl_screen_position(200,200));
//			if(Frame==(SeqLength/2))
//			MacVideo.display_image(ByteImage, mvl_screen_position(200+XSize+1,200));
//			if(Frame==(SeqLength-1))
//			MacVideo.display_image(ByteImage, mvl_screen_position(200+2*XSize+2,200));

			} // end for
	
	float Time = Timer.time();
	
	// reset hyperplane size
	ByteImage.set_hyperplane_size(3);
	return(Time);
} // end drifter
