/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include "iml_poly.h"
#include "iml_err.h"

#include "stdio.h"

// JUNK RETURN VALUE FOR ERROR CASES
IML_DOUBLE polynomial::X = 1;

	
//==========================================================================
// polynomial constructor
//==========================================================================
polynomial::polynomial(polynomial IML_REF Poly) {

	X = Accumulator = 0.0;
   Order = Poly.Order;
	Coefficients = new IML_DOUBLE[Order+1];
	if(!Coefficients) {
		iml_error("polynomial", "Could not allocate temporary space!  Init failed...");
		return;
		} // end if
	for(IML_USINT I=0;I<=Order;I++) {
		Coefficients[I]=Poly.Coefficients[I];
		} // end for
};

//==========================================================================
// polynomial constructor
//==========================================================================
polynomial::polynomial(IML_ULINT POrder) {

	X = Accumulator = 0.0;
   Order = POrder;
	Coefficients = new IML_DOUBLE[Order+1];
	if(!Coefficients) {
		iml_error("polynomial", "Could not allocate temporary space!  Init failed...");
		return;
		} // end if

	// INITIALIZE ALL TO ZERO
	for(IML_USINT I = 0; I <= POrder; I++) {
		Coefficients[I] = 1.0;
		} // end for

} // end polynomial::polynomial


//==========================================================================
// polynomial destructor
//==========================================================================
polynomial::~polynomial(void) {
	// RELEASE THE MEMORY
	if(Coefficients) delete [] Coefficients;

} // end polynomial::~polynomial


//==========================================================================
// coefficient value accessor/setter
//==========================================================================
IML_DOUBLE IML_REF polynomial::operator[](IML_ULINT CoeffIndex) {

	if( CoeffIndex > Order ) {
		iml_error("polynomial","Out of bounds coefficient index");
		return(X); // undefined junk value
		} // end if

	return(Coefficients[CoeffIndex]);

} // end polynomial::operator[]


//==========================================================================
// virtual polynomial evaluator; derived classes can evaluate
//==========================================================================
IML_DOUBLE polynomial::operator()(IML_DOUBLE XVal) {

	X = XVal;
	Accumulator = Coefficients[Order];

	for(IML_LINT P = Order-1; P >= 0; P--) {
		Accumulator *= X;
		Accumulator += Coefficients[P];
		} // end for

	return(Accumulator);
} // end polynomial::operator()




//==========================================================================
// polynomial constructor
//==========================================================================
hermite_polynomial::hermite_polynomial(
		IML_ULINT POrder, IML_DOUBLE Sigma) : polynomial(POrder) {

	// SET UP ODD-EVEN FLAG
	OddOrder = POrder % 2;
	Konst = -2.0/(Sigma*Sigma);

	// GENERATE THE HERMITE POLYNOMIAL COEFFICIENTS FOR ORDER 1 OR 0
	if(POrder == 0) {
		Coefficients[0] = 1.0;
      return;
	} else if(POrder == 1) {
		Coefficients[1] = Konst;
      return;
	} // end if

		// OTHERWISE WE HAVE TO GENERATE THEM VIA THE
		// RECURSION RELATION: H(n) = x * H(n-1) - (n-1) * H(n-2)
		//
		// eric: describe the method here!!!
      //
	IML_DOUBLE * GenN   = new IML_DOUBLE[POrder+1];	// ALLOCATE TEMP SPACE
	IML_DOUBLE * GenNm1 = new IML_DOUBLE[POrder+1];	// ALLOCATE TEMP SPACE
	IML_DOUBLE * GenNm2 = new IML_DOUBLE[POrder+1];	// ALLOCATE TEMP SPACE

	if(!(GenN && GenNm1 && GenNm2 && Coefficients)) {
		iml_error("hermite_polynomial", "Could not allocate temporary space!  Init failed...");
		// CLEAN UP
		if(GenN)   delete [] GenN;
		if(GenNm1) delete [] GenNm1;
		if(GenNm2) delete [] GenNm2;
		return;
		} // end if

	IML_SINT I,J;
	// INITIALIZE TEMP SPACE
	for(I = 0; I <= POrder; I++)  GenN[I] = GenNm1[I] = GenNm2[I] = 0.0;

	// INITIALIZE THE RECURSION RELATION
	GenNm1[1] = Konst;
	GenNm2[0] = 1.0;

	// I HAVE SERIALIZED THE RECURSION RELATION :-)
   IML_DOUBLE *Temp;   // TEMP FOR SWAPPING AROUND THE ARRAYS
	for(I = 2; I <= POrder; I++) { // GENERATE SUCCESSIVE Hn COEFF ARRAYS

		int NFac = I-1;

   	for(J = I; J > 0; J--) {
			GenN[J] = Konst*GenNm1[J-1] + Konst*NFac*GenNm2[J];
			} // end for
		GenN[0] = Konst*NFac*GenNm2[0];

		// SWAP ARRAYS
		Temp 		= GenNm2;
		GenNm2 	= GenNm1;
		GenNm1 	= GenN;
		GenN 		= Temp;
		} // end for

	// STORE GENERATED COEFFICIENTS
	for(I=0;I<=POrder;I++)  Coefficients[I] = GenNm1[I]; // INITIALIZE REAL ARRAY

	// CLEAN UP
	delete [] GenN;
	delete [] GenNm1;
	delete [] GenNm2;
} // end hermite_polynomial::polynomial


//==========================================================================
// virtual hermite_polynomial evaluator
//==========================================================================
IML_DOUBLE hermite_polynomial::operator()(IML_DOUBLE XVal) {

	X = XVal*XVal;
	Accumulator = Coefficients[Order];

	// HERMITE POLYNOMIALS USE ONLY EITHER ODD OR EVEN POWER TERMS
	for(IML_LINT P = Order-2; P >= 0; P-=2) {
		Accumulator *= X;
		Accumulator += Coefficients[P];
//fprintf(stderr,"partial<%f>\n",(IML_FLOAT)Accumulator);
		} // end for

	if(OddOrder) Accumulator *= XVal;

	return(Accumulator);
} // end polynomial::operator()


