#ifndef __IML_POLYNOMIALS__
#define __IML_POLYNOMIALS__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include "iml_type.h"
#include "iml_math.h"

/** This class implements efficient evaluation of uni-variate polynomials.
By efficient, I mean the method of evaluation requiring the fewest computer
instructions for a polynomial with arbiratrary order and coefficient values.
The evaluation method for an Nth order polynomial is:
<pre>

	a0 + x(a1 + x(a2 + ...x*aN)...)

</pre>

No exotic capabilities are included (like symbolic differentiation, although that
would be easy to add) because I'm sure that someone else has already done
it, whatever "it" happens to be.  And they've probably done it in the "right"
too. :-)
*/
class polynomial {

protected:
	static IML_DOUBLE X; // USE FOR RETURNING DEFAULT NONSENSE VALUE.
	IML_LDOUBLE Accumulator;
	IML_ULINT Order;
	IML_DOUBLE IML_PTR Coefficients;

public:
	/// Allocate a polynomial of order POrder
	polynomial(IML_ULINT POrder);
	
	/// Duplicate another polynomial.
	polynomial( polynomial IML_REF Poly );

	
	/// Make sure to deallocate memory.
	~polynomial(void);
	
	/** This function provides access to the polynomial
	 coefficients.  The index is bounds checked. Use this to fill
	 in the coefficient values.  The constructor sets all coefficients
	 to 1.0  
	 */
	IML_DOUBLE IML_REF operator[](IML_ULINT CoeffIndex);
	
	/** This function evaluates the polynomial to double precision.  
	*/
	IML_DOUBLE operator()(IML_DOUBLE XVal);

}; // end class polynomial


/** This class implements a special subset of polynomials.  This class
is why I created polynomials in the first place. I wanted
derivative of Gaussians, and the Nth derivative of a Gaussian is just the
Nth order Hermite multiplied by the original Gaussian.  Cool, huh? :-)
I <em>did</em> go to the trouble of serializing the recursive definition

<pre>

	H(n) = x * H(n-1) - (n-1) * H(n-2)

</pre>

of the Hermite in order to avoid estimating required stack space. That was
fun. ;-)
*/
class hermite_polynomial : public polynomial {
protected:
	IML_LINT OddOrder;
	IML_DOUBLE Konst;
public:
	/** The only constructor.  You must specify the order, with optional
	specification of a scaling constant.  You need the scaling constant
	if you are using this for derivative of a Guassian.  That constant depends on
	both the standard deviation and the order of the derivative, and is
	<pre>
	
		(sqrt(2)/SDev)**POrder
	
	</pre>
	where DerivativeOrder == POrder and '**' denotes exponentiation.
	*/
	hermite_polynomial(IML_ULINT POrder, IML_DOUBLE Sigma = IML_SQRT2);
	
	/** Hide the other evaluation method.  This can be about twice as fast as 
	the standard evaluation method because Hermites have only even or odd
	terms in the polynomial. 
	*/
	IML_DOUBLE operator()(IML_DOUBLE XVal);


}; // end class hermite_polynomial

#endif // __IML_POLYNOMIALS__
