#ifndef __IMLRAND_H__
#define __IMLRAND_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

// THE FOLLOWING ARE R.E. FREDERICKSEN'S ATTEMPT AT
// PROVIDING A SET OF RANDOM NUMBER GENERATOR CLASSES
// WITH GOOD PROPERTIES.


#include "iml_err.h"
#include "iml_type.h"
#include "iml_math.h"

#include <math.h>
#include <time.h>


// HERE IS AN ABSTRACT BASE CLASS SO WE CAN USE GENERIC
// RANDOM VARIABLE "pointers" IN OUR CODE...

/** This is the base class for scalar random variable objects. 
It provides virtual functions which define an interface for 
talking to random variables.
*/
class scalar_rv {

//		long int SeedVal;

	public: 
	
   	/// This is the (default) constructor.
		scalar_rv(void) { /*seed(-1);*/ };

		/// This function is used to set the random variable seed.
		virtual void seed(long Seed)=0;// { /*SeedVal = Seed;*/ }; 
			
		/** The function must be over-ridden in derived classes.
		Draw a sample from the the distribution.
		*/
		virtual double sample(void) { return(0.0); }

		/** This function must be over-ridden by a derived class. This function 
		should be used to configure the distribution's mean and standard deviation. 
		Higher moment configuration might be needed for some distributions
		but that is not handled by the current class design.
		*/
		virtual void reset(float,float) {};

}; // end class scalar_rv


/** This is the base class for uniform random variable objects.
It is intended to be used to generate other distributions.  
The base class provides a one-member-function interface for one or more functions.    
Any public domain random number generator with
decent properties will do.
*/
class unit_uniform_rv {
	public:

		/** The function must be over-ridden in derived classes.
		Draws a sample from the distribution.
		*/
		virtual double sample(void) { return(0.0); };

		/// This function is used to set the random variable seed.
		virtual void seed(long Seed) {  }; 

}; // end class


static const long matsumoto_N = 624;
/** This class encapsulates Makoto Matsumoto's MERSENNE TWISTER 
random number generator.   The user need not know anything about how it works.
I found this on a neat web site that is all about 
<A HREF = "http://random.mat.sbg.ac.at/"> random numbers. </A>
<p> <b> Official Blurb: </b> </p>
 <td width=70%>
    Creating a generator with period 2**800 was not enough for
    <A HREF="mailto:matumoto@math.keio.ac.jp">Makoto Matsumoto</A>. 
    The Mersenne Twister, a new variant of TGFSR by Matsumoto and Nishimura,
    sets new standards for the period, quality and speed of random number generators.
    The incredible period is 2**19937 - 1, a number with about 6000 decimal
    digits; the 32-bit random numbers exhibit best possible equidistribution
    properties in dimensions up to 623; and it's fast, very fast.
    But even on a DEC Alpha Station, the period is so huge that producing all the numbers
    would require about 10**5983 millenia. The 
    <A HREF="ftp://random.mat.sbg.ac.at/pub/data/mt19937b-int.c">
    integer version</A>
    and the 
    <A HREF="ftp://random.mat.sbg.ac.at/pub/data/mt19937b-real.c">
    floating point version</A>
    are available in C.
    A paper on the Mersenne Twister is submitted to ACM TOMACS.
    </td>
 <p></p>*/
class matsumoto_MT19937B : public unit_uniform_rv {
//		static const long matsumoto_N;
		unsigned long ptgfsr[matsumoto_N]; /* set initial seeds: N = 624 words */
		unsigned long Ctr;

		void sgenrand(unsigned long Seed);
		double genrand(void);
		
	public:
		/** The only constructor.*/
		matsumoto_MT19937B(void) { Ctr = matsumoto_N; sgenrand(4357); };
		/** Correctly overload the sample() function.
		The function must return a value in the range 0 to 1.
		*/
		virtual double sample(void);
		/** Provide the seed() function.*/
		virtual void seed(long Seed) { sgenrand(Seed); };

}; // end class


static const long FSULTRA_N=37;          /* size of table        */
//========================================================================
/** This is the Swiss Army Knife approach to random number generators.
Here is the <b>Official Blurb</b>, although I've adapted it to be a class
in this library.  Read the code for how this works; in brief, it mixes
a congruential RNG with a shift-register RNG to produce the final output.
<pre>

FSU - ULTRA	The greatest random number generator that ever was
		or ever will be.  Way beyond Super-Duper.
		(Just kidding, but we think its a good one.)

Authors:	Arif Zaman (arif@stat.fsu.edu) and
		George Marsaglia (geo@stat.fsu.edu).

Date:		27 May 1992

Version:	1.05

Copyright:	To obtain permission to incorporate this program into
		any commercial product, please contact the authors at
		the e-mail address given above or at

		Department of Statistics and
		Supercomputer Computations Research Institute
		Florida State University
		Tallahassee, FL 32306.

See Also:	README		for a brief description
		ULTRA.DOC	for a detailed description
</pre>
*/
//========================================================================
class fsultra : public unit_uniform_rv {

		long swb32[FSULTRA_N];		// These arrays contian the seed mixed with
		long swb16[FSULTRA_N]; 		// a congruential. They are used 32, 16, and 8
		long swb8[FSULTRA_N]; 		// bits at a time respectively.
		long swbseed[FSULTRA_N];

		char	flags;		/* The carry flag for the SWB generator	*/
		long unsigned congx;    /* Seed for congruential generator	*/

		long	*swb32p;	short	swb32n;	/* A counter and a	*/
		short	*swb16p;	short	swb16n;	/* pointer is kept for	*/
		char	*swb8p;		short	swb8n;	/* each array.		*/

		char     swb1[32];				/* The swb1 is slightly	*/
		char    *swb1p;		short	swb1n;	/* different.		*/

				
		void	rinit(long unsigned congy,long unsigned shrgx);
		void	SWBfill(long *x);
		long	swb32fill(void);
		short	swb16fill(void);
		char	swb8fill(void);
		char	swb1fill(void);
		unsigned long	bit_reverse( unsigned long n );
		
	public:
		/** The only constructor.*/
		fsultra(void);
		/** Correctly overload the sample() function.
		The function must return a value in the range 0 to 1.
		*/
		virtual double sample(void){ return(duni()); };
		/** Correctly overload the seed() function.  We have to do a 
		little extra here, because rinit() wants two seeds, each to 
		be used differently.  I just do a bit reversal on the first seed
		to create the second seed.
		*/
		virtual void seed(long Seed) { rinit( Seed, bit_reverse(Seed) ); };

		/** @name Other blades on the Swiss Army Knife. */
		//@{
		/** Signed 32 bit number. */
		long  i32bit(void) { return  (swb32n--) ? *(swb32p++) : swb32fill(); };
		/** Unsigned 31 bit number. */
		long  i31bit(void) { return ((swb32n--) ? *(swb32p++) : swb32fill()) & 0x7FFFFFFF; };
		/** Signed 16 bit number. */
		short i16bit(void) { return  (swb16n--) ? *(swb16p++) : swb16fill(); };
		/** Unsigned 15 bit number. */
		short i15bit(void) { return ((swb16n--) ? *(swb16p++) : swb16fill()) & 0x7FFF; };
		/** Signed 8 bit number. */
		char  i8bit(void)  { return  (swb8n--)  ? *(swb8p++)  : swb8fill(); };
		/** Unsigned 7 bit number. */
		char  i7bit(void)  { return ((swb8n--)  ? *(swb8p++)  : swb8fill())  & 0x7F; };
		/** Random binary number. */
		char  i1bit(void)  { return  (swb1n--)  ? *(swb1p++)  : swb1fill(); };

		/** Random 32-bit floating point in the range -1 to 1. */
		float vni(void);
		/** Random 32-bit floating point in the range 0 to 1. */
		float uni(void);
		/** Random 64-bit floating point in the range -1 to 1. */
		double dvni(void);
		/** Random 64-bit floating point in the range 0 to 1. */
		double duni(void);
		//@}
}; // end class


/** This class produces the interface for a user-accesible uniform random variable. 
The class contains a pointer to a unit_uniform_rv object,
plus instances of each type of specific unit_uniform_rv (child classes
derived from unit_uniform_rv).  Configuring use of any specific type requires a 
member function call. The default constructor produces a uniform random variable 
over (0, 0.99999999999999) But the object can be configured to have any endpoints representable by single 
precision floating point.
<p></p>*/
class uniform_rv : public scalar_rv {

	private:  // COMPLETELY PRIVATE STUFF GOES HERE
		float MinR;
		float MaxR;
		float Range;


	protected: // STUFF THAT CAN BE INHERITED PUBLICLY GOES HERE
		// FOUR DOCUMENTED "GOOD" UNIFORM
		// RANDOM VARIABLE GENERATORS
		unit_uniform_rv *UU;
		matsumoto_MT19937B MT19937B;
		fsultra	FSULTRA;

	public: // STUFF THAT IS ALWAYS PUBLIC GOES HERE

   	/// We only a need a single constructor.
		uniform_rv(float Min = 1e-30, float Max = 1 - 1e-30) : scalar_rv() {
			MinR = Min; MaxR = Max; Range = Max-Min;
			use_MT19937B();
			};

		/** Select Matsumoto's MERSENNE TWISTER unit-uniform random number generator. */
		void use_MT19937B(void) { UU = &MT19937B; }

		/** Select the FSULTRA unit-uniform random number generator. */
		void use_FSULTRA(void) { UU = &FSULTRA; }

		/** Returns a value in the range 0 to 1.  */
		virtual void seed(long Seed) { UU->seed(Seed); }; 

#define IML_SQRT3	1.732050808f

		/** Here is the required over-ride for the reset() interface function.
		It configures the mean and standard deviation of the uniform random varible.
		These numbers are translated into distribution end points.
		<p></p>*/
		void reset(float Mean, float SDev) {
			MinR = Mean - SDev*IML_SQRT3;
			MaxR = Mean + SDev*IML_SQRT3;
			Range = MaxR - MinR;
			}

		/** Set the distribution end points.  
		This is an alternative to setting the mean and standard deviation, 
		since that is less intuitive for a uniform distribution.
		*/
		void set_min_max(float Min, float Max)
			{ MinR = Min; MaxR = Max; Range = Max-Min;}

		/** Here is the required over-ride for the sample() interface function.
		*/
		double sample(void) { return( MinR + Range * UU->sample()); }

}; // end class uniform_rv


/** The class defines a binary-random-number generator.  
The interface for using it is defined in the scalar_rv class: reset() and sample().
*/
class binary_rv : public uniform_rv {

	private:
	float Value1;
   	float Value2;

	public:
		/** The only constructor. 
		Creates a binary random variable generating values of {V1,V2} 
		with equal probability.
		*/
		binary_rv(float V1 = 0, float V2 = 1.0) : uniform_rv(1e-30f, 1.0) {
			Value1 = V1;  Value2 = V2;
			};

		/** Here is the required over-ride for the reset() interface function.
		*/
		void reset(float MeanV, float SDevV) {
			Value1 = MeanV-SDevV;	Value2 = MeanV+SDevV;  
			};

		/** Here is the required over-ride for the sample() interface function.
		*/
			// WE KNOW THAT uniform_rv::sample() RETURNS A NON-ZERO NUMBER, BY DEFAULT
		double sample(void) {
		
			//return( FSULTRA.i1bit() > 0.5 ? Value1 : Value2 ); 
			return( uniform_rv::sample() > 0.5 ? Value1 : Value2 ); 
			};

}; // end class binary_rv


/** The class defines a Gaussian-random-number generator.  
The interface for using it is defined in the scalar_rv class: reset() and sample().
*/
class gaussian_rv : public uniform_rv {

	private:
	float Mean;
   	float SDev;

	public:
		/** The only constructor. 
		Defaults to a normal distribution.
		*/
		gaussian_rv(float MeanV = 0, float SDevV = 1.0) : uniform_rv(1e-30f, 1.0) {
			Mean = MeanV;  SDev = (SDevV > 0) ? SDevV : -SDevV;
			};

		/** Here is the required over-ride for the reset() interface function.
		*/
		void reset(float MeanV, float SDevV) {
			Mean = MeanV;  SDev = (SDevV > 0) ? SDevV : -SDevV;
			};

		/** Here is the required over-ride for the sample() interface function.
		*/
			// WE KNOW THAT uniform_rv::sample() RETURNS A NON-ZERO NUMBER, BY DEFAULT
		double sample(void) {
			double register temp = SDev * sqrt( -2 * log( uniform_rv::sample() ) );
//			return( Mean + temp * (UU->rand() > 0.5 ? -1 : 1) ); 
			return( Mean + temp * sin(uniform_rv::sample()*IML_2PI) ); 
			};

}; // end class gaussian_rv


/** The class defines an Exponential-random-number generator.  
The interface for using it is defined in the scalar_rv class: reset() and sample().
*/
class exponential_rv : public uniform_rv {

   // JUST REPLACE THE rand() FUNCTION WITH THE EXPDEV VERSION
	private:
		double Lambda;

	public:

		/// The single constructor; defines a default value.
		exponential_rv(float LLambduh = 1.0) : uniform_rv(1e-30f, 1.0) {
			Lambda = (LLambduh > 0.0) ? LLambduh : 1.0;
			};

//  DISABLE THE WARNING SO AS NOT TO ALARM PEOPLE.
//MACMOD -- REF oct/14/96
//#pragma warn -hid  // TURN OFF HIDDEN VIRTUAL FUNCTION WARNING

		/** Here is the required over-ride for the reset() interface function.
		The reset function has only one parameter (the mean) while other 
		distributions can have two (mean, sdev).
		*/
		virtual void reset(float LLambduh, float Dummy=0) {
			Lambda = (LLambduh > 0.0) ? LLambduh : 1.0;
			};
			
//MACMOD -- REF oct/14/96
//#pragma warn .hid  // RETURN IT TO ITS PREVIOUS STATE

		/// Here we correctly over-ride the sample() interface function.
		double sample(void) {
			// WE KNOW THAT uniform_rv::sample() RETURNS A NON-ZERO NUMBER, BY DEFAULT
			return(-log(uniform_rv::sample()) * Lambda); 
			};

}; // end class exponential_rv



#endif // __IMLRAND_H__
