#ifndef __IML_ROFN_H__
#define __IML_ROFN_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1993-99 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include "iml_type.h"
#include "iml_domn.h"
#include "iml_err.h"
#include "iml_poly.h"
#include "iml_math.h"
#include "iml_xfrm.h"



//=========================================================================
// AN ABSTRACT 2D FUNCTION GENERATOR CLASS. IN THIS CLASS WE PROVIDE
// FOR SCALING, TRANSLATION OF THE FUNCTIONS.
//=========================================================================
// A READABLE BASE CLASS
/** This class provides the interface for <em>read-only functions</em>.
The idea is that n-dimensional images can be filled by n-dimensional
functions by overloading the assignment operator.
</p> */
class iml_rofn : public iml_domain, public iml_xlate_scale {

	protected:

		IML_DOUBLE Magnitude;
		friend class iml_image;
		
	public:
		iml_rofn(IML_ULINT DimCnt) : iml_domain(DimCnt), iml_xlate_scale(DimCnt) 
			{ set_magnitude(1.0); };
		~iml_rofn() {};

		void set_magnitude(IML_DOUBLE Mag) { Magnitude = Mag; };

		IML_DOUBLE get_magnitude(void) {return(Magnitude); };

		//=======================================================================
		/** One-dimensional index for ease of use. If the read only function
		is of other dimension the return value is undefined.
		</p> */
		IML_DOUBLE operator()(IML_DOUBLE X0) {
			Domain.set_ns(IML_2D_X, X0); return(get_indexed_value());
			};
		/** Two-dimensional index for ease of use. If the read only function
		is of other dimension the return value is undefined.
		</p> */
		IML_DOUBLE operator()(IML_DOUBLE X0, IML_DOUBLE X1) {
			Domain.set_ns(IML_2D_X, X0); Domain.set_ns(IML_2D_Y, X1); 
			return(get_indexed_value());
			};
		/** Three-dimensional index for ease of use. If the read only function
		is of other dimension the return value is undefined.
		</p> */
		IML_DOUBLE operator()(IML_DOUBLE X0, IML_DOUBLE X1, IML_DOUBLE X2) {
			Domain.set_ns(0, X0); Domain.set_ns(1, X1); Domain.set_ns(2, X2); 
			return(get_indexed_value());
			};
		//=======================================================================
		// FOR CONVENIENCE, ADD A VARIABLE ARGUMENT LIST ACCESSOR FOR 
		// SIMPLIFIED SYNTACTICAL ACCESS. I.e. YOU DON'T HAVE TO USE
		// INDEX OBJECTS . . .
		//=======================================================================
		/** Four-or-more-dimensional access for ease of use. If the read only function
		is of other dimension the return value is undefined.
		</p> */
		IML_DOUBLE operator()(IML_DOUBLE X0, IML_DOUBLE X1, IML_DOUBLE X2, ...);

		/** Access the function using an index object. If the read only function
		is of other dimension the return value is undefined.
		</p> */
		IML_DOUBLE get_value( iml_index IML_REF Index);

		POLYMORPHIC IML_DOUBLE get_indexed_value(void) {
			iml_error("iml_rofn", "Call to get_indexed_value: should be replaced.");
			return(1.0);
			}
			
}; // end class iml_rofn



//=====================================================================
/** An <em>inverse-distance</em> function.  This function generates
 the L-Norm of the
distance from the pixel (it's position in the space) to the origin.
<pre>

			    Magnitude                     
       ------------------------------  = 
      (1.0 + Magnitude * LNORM(X))  

 where X is the input vector.

</pre>
</p> */
class inverse_distance_rofn : public iml_rofn {

	public:
		IML_DOUBLE LNorm;

		inverse_distance_rofn(IML_ULINT DimCnt = 2, IML_DOUBLE LNormV = 2.0) 
				: iml_rofn(DimCnt) 
			{ set_lnorm(LNormV); };
		
		// copy constructor, just in case
		inverse_distance_rofn(inverse_distance_rofn IML_REF IDrofn) 
				: iml_rofn(IDrofn.Domain.Dimensionality) 
			{ set_lnorm(IDrofn.LNorm); };
			
		~inverse_distance_rofn() {};

		void set_lnorm(IML_DOUBLE LNormV);
			
		IML_DOUBLE get_indexed_value( void ) ;

}; // end class inverse_distance_rofn


//=========================================================================
/** This is a base class for trigonometric style functions.
</p> */
//=========================================================================
class iml_trig_rofn	: public iml_rofn {

	protected:
		IML_DOUBLE *Freq;
		IML_DOUBLE Phase;
		
		void initialize(void);

	public:

		iml_trig_rofn(IML_ULINT DimCnt = 2);
		
		// copy constructor, just in case
		iml_trig_rofn(iml_trig_rofn IML_REF TFrofn);
			
		~iml_trig_rofn();

		/** Set the period (in pixels) of the sinusoid in the requested dimension.
		A period of zero signals infinate period,
		or zero frequency. See also <em>set_frequency()</em>. 
		</p> */
		void set_period(IML_ULINT Dim, IML_DOUBLE Period) {
			if( Period == 0 ) Freq[Dim] = 0.0;
			else { Freq[Dim] = 1.0/Period; }
      	};

		/** Set the frequency (in cycles/pixel) of the sinusoid in the requested dimension.
		See also <em>set_period()</em>.
		</p> */
		void set_frequency(IML_ULINT Dim, IML_DOUBLE Fr) { Freq[Dim] = Fr; };

		/** Find out the current frequency. </p> */
		IML_DOUBLE get_frequency(IML_ULINT Dim) { return(Freq[Dim]); };

		/** Set the phase of the sinusoid, in degrees. </p> */
		POLYMORPHIC void set_phase(IML_DOUBLE Ph) { Phase = Ph*IML_RAD_PER_DEG; };
		/** Get the phase of the sinusoid, in degrees. </p> */
		POLYMORPHIC IML_DOUBLE get_phase(void) { return(Phase*IML_DEG_PER_RAD); };

}; // end class iml_trig_rofn


//=====================================================================
/** An n-dimensional cosine. This function generates the value:
<p>
<pre>

	Magnitude * cos( 2*pi*F.X + P)

</pre>
<p>
where X us the pixel position (vector), F is the frequency vector, in cycles/pixel, 
and P is the phase.
</p> */
class iml_cos_rofn : public iml_trig_rofn {
friend class iml_gabor_rofn;
	public:
		iml_cos_rofn(IML_ULINT DimCnt = 2) : iml_trig_rofn(DimCnt) { };
		
		// copy constructor, just in case
		iml_cos_rofn(iml_cos_rofn IML_REF COSrofn) 
			: iml_trig_rofn(COSrofn.Domain.Dimensionality) {};
			
		~iml_cos_rofn() {};

		IML_DOUBLE get_indexed_value( void );

}; // end class iml_cos_rofn


//=====================================================================
/** A polar cosine.  This really only makes sense in 2-D or 3-D. This function 
generates the value:
<p>
<pre>

	Magnitude * cos( 2*pi*((Fr, Ft, Fp).(R,THETA,PHI)) + P)
	
</pre>
<p>
where R, THETA and PHI are the pixel position in polar coordinates,
Fr,Fr and Fp are the corresponding frequencies, and P is the phase.
</p> */
class iml_polar_cos_rofn : public iml_trig_rofn, public iml_polarizer {
	public:
		iml_polar_cos_rofn( IML_ULINT DimCnt = 2) : iml_trig_rofn(DimCnt) {};
		
		// copy constructor, just in case
		iml_polar_cos_rofn(iml_polar_cos_rofn IML_REF PCOSrofn) 
			: iml_trig_rofn(PCOSrofn.Domain.Dimensionality) {};

		~iml_polar_cos_rofn() {};

		IML_DOUBLE get_indexed_value( void );

}; // end class iml_polar_cos_rofn




//=====================================================================
// THIS FUNCTION GENERATES THE VALUE
//					      [ -(x-mx)^2     -(y-my)^2]
//  Magnitude *  exp [  --------  +  ---------]  and  rotated Phi degrees
//                   [    sx^2          sy^2  ]
//=====================================================================
/** An n-dimensional Gaussian.  This function generates the value:
<p>
<pre>
                   [ -(x1-m1)^2        -(xn-mn)^2]
  Magnitude *  exp [  --------  + . . + ---------]
                   [    s1^2              sn^2   ]
</pre>
<p>
where the <em>mi</em> are the <em>n</em> means, 
and <em>si</em> are the <em>n</em> space constants.
<p> */
class iml_gaussian_rofn : public iml_rofn {

	public:

		iml_gaussian_rofn(IML_ULINT DimCnt = 2) : iml_rofn(DimCnt) {	};
		
		// copy constructor just in case
		iml_gaussian_rofn(iml_gaussian_rofn IML_REF G) : iml_rofn(G.Domain.Dimensionality) {	};
		
		~iml_gaussian_rofn() {};
		
		/** Set the clipping of the dimension to some multiple of standard
		deviations.  The default is 3.0.  This saves time when assigning a
		small gaussian to a large image.
		<p> */
		void set_domain_sd(IML_ULINT Dim, IML_DOUBLE NSDev=3.0) {
			IML_DOUBLE M = get_center(Dim);
			IML_DOUBLE SD = get_sdev(Dim);
			IML_DOUBLE Del = NSDev*SD;
			Domain.set_min_max(Dim, M-Del, M+Del);
			};

		/** Set the mean of the gaussian in the
		requested dimension (in pixel coords).
		<p> */
		void set_mean(IML_ULINT Dim, IML_DOUBLE Mean) { set_center(Dim,Mean); };

		/** Get the mean of the gaussian in the
		requested dimension (in pixel coords).
		<p> */
		IML_DOUBLE get_mean(IML_ULINT Dim) { return(get_center(Dim)); };

		/** Set the standard deviation of the gaussian in the
		requested dimension (in pixels).
		<p> */
		void set_sdev(IML_ULINT Dim, IML_DOUBLE SDev) { set_scale(Dim,SDev); };

		/** Get the standard deviation of the gaussian in the
		requested dimension (in pixels).
		<p> */
		IML_DOUBLE get_sdev(IML_ULINT Dim) { return(get_scale(Dim)); };

		IML_DOUBLE get_indexed_value( void );

}; // end class iml_gaussian_rofn


//=====================================================================
/** This function is an n-dimensional gaussian multplied by an n-dimensional
cosine.  The cosine is a member the gaussian, and is configured
directly; see the <em>iml_cos_rofn</em> class.  See the
see the <em>iml_gaussian_rofn</em> class for configuration of the
gaussian.
</p> */
//=====================================================================
class iml_gabor_rofn : public iml_gaussian_rofn {

	public:
		/** The sinusoidal carrier; see the <em>iml_cos_rofn</em> class.
		</p> */
		iml_cos_rofn Carrier;

		iml_gabor_rofn(IML_ULINT DimCnt = 2) : 
			iml_gaussian_rofn(DimCnt),  Carrier(DimCnt) {};
				
		// copy constructor just in case
		iml_gabor_rofn(iml_gabor_rofn IML_REF G) : 
			iml_gaussian_rofn(G.Domain.Dimensionality),  Carrier(G.Domain.Dimensionality) {};

		~iml_gabor_rofn() {};

		IML_DOUBLE get_indexed_value( void );

}; // end class iml_gabor_rofn


//=====================================================================
// THIS FUNCTION GENERATES THE VALUE
//					      [ -(x-m)^2   ]
//  Magnitude *  exp [  --------  ]
//                   [    S       ]
//=====================================================================
class iml_symmetric_gaussian_rofn : public iml_gaussian_rofn {
	public:
		iml_symmetric_gaussian_rofn(IML_ULINT DimCnt = 2) : iml_gaussian_rofn(2) {};

		// copy constructor just in case
		iml_symmetric_gaussian_rofn(iml_symmetric_gaussian_rofn IML_REF G) 
			:	iml_gaussian_rofn(G.Domain.Dimensionality) {};

		~iml_symmetric_gaussian_rofn() {};

		void set_sdev(IML_DOUBLE SDev);
		IML_DOUBLE get_sdev(void) { return(get_scale(IML_2D_X)); };

}; // end class iml_symmetric_gaussian_rofn


//=====================================================================
//
//=====================================================================
class iml_raised_gaussian_rofn : public iml_gaussian_rofn {
	IML_DOUBLE *Radii;

	public:
		iml_raised_gaussian_rofn(IML_ULINT DimCnt = 2) : iml_gaussian_rofn(DimCnt) {
			Radii = new IML_DOUBLE[DimCnt];
			if( !Radii ) iml_error("iml_raised_gaussian_rofn","Can't allocate array.");
			};
			
		iml_raised_gaussian_rofn(iml_raised_gaussian_rofn IML_REF RG) 
				: iml_gaussian_rofn(RG.Domain.Dimensionality) {
			Radii = new IML_DOUBLE[RG.Domain.Dimensionality];
			if( !Radii ) iml_error("iml_raised_gaussian_rofn","Can't allocate array.");
			};
			
			
		~iml_raised_gaussian_rofn() { if(Radii) delete [] Radii; };

		void set_radius(IML_ULINT Dim, IML_DOUBLE Rad)	{ Radii[Dim] = Rad; };
		IML_DOUBLE get_radius(IML_ULINT Dim)	{ return(Radii[Dim]); };
                                          
		IML_DOUBLE get_indexed_value( void );

}; // end class iml_raised_gaussian_rofn


//=====================================================================
//
//=====================================================================
class iml_raised_cos_rofn : public iml_cos_rofn {
	IML_DOUBLE *Radii;

	public:

		iml_raised_cos_rofn(IML_ULINT DimCnt = 2) : iml_cos_rofn(DimCnt) {
			Radii = new IML_DOUBLE[DimCnt];
			if( !Radii ) iml_error("iml_raised_gaussian_rofn","Can't allocate array.");
			};
			
		iml_raised_cos_rofn(iml_raised_cos_rofn IML_REF RC) : iml_cos_rofn(RC.Domain.Dimensionality) {
			Radii = new IML_DOUBLE[RC.Domain.Dimensionality];
			if( !Radii ) iml_error("iml_raised_gaussian_rofn","Can't allocate array.");
			};
			
		~iml_raised_cos_rofn() { if(Radii) delete [] Radii; };

		void set_radius(IML_ULINT Dim, IML_DOUBLE Rad)	{ Radii[Dim] = Rad; };
		IML_DOUBLE get_radius(IML_ULINT Dim)	{ return(Radii[Dim]); };

                                          
		IML_DOUBLE get_indexed_value( void );

}; // end class iml_raised_cos_rofn



//=====================================================================
// THIS FUNCTION GENERATES THE VALUE
//					     [ -(X-M)^2  ]
//  Magnitude * exp [  --------  ] * Hx
//                  [    S^2    ]
//
// where Hx is the appropriate Hermite polynomial
//=====================================================================
class iml_dgaussian_rofn : public iml_gaussian_rofn {

// LOCALLY USED FUNCTION
	IML_DOUBLE calc_derivative_const(IML_DOUBLE SD, IML_ULINT DOrder);

	IML_ULINT *Order;
	IML_DOUBLE *DScaler;
	hermite_polynomial **Poly;
	void initialize(void);

	public:

		iml_dgaussian_rofn( IML_ULINT DimCnt = 2 );
		iml_dgaussian_rofn( iml_dgaussian_rofn IML_REF DG );
		~iml_dgaussian_rofn();

		void set_order(IML_ULINT Dim, IML_ULINT Der);
		void set_sdev(IML_ULINT Dim, IML_DOUBLE SDev);
		IML_DOUBLE get_indexed_value( void );

}; // end class iml_dgaussian_rofn


//=====================================================================
// THIS FUNCTION GENERATES THE VALUE
//  Magnitude * rect( (X-D)/T)
//     WHERE rect( X ) is defined as
//				{  Magnitude  if (|X-D| <= 1/2) 
// 			{  0          otherwise
//=====================================================================
class iml_rect_rofn : public iml_rofn {
	public:
		iml_rect_rofn(	IML_ULINT DimCnt = 2 ) : iml_rofn(DimCnt) { };
		iml_rect_rofn(	iml_rect_rofn IML_REF R ) : iml_rofn(R.Domain.Dimensionality) { };
		~iml_rect_rofn() {};

		void set_width(IML_ULINT Dim, IML_DOUBLE Width) { set_scale(Dim, Width/2.0); }; 
		void set_halfwidth(IML_ULINT Dim, IML_DOUBLE HWidth) { set_scale(Dim, HWidth); }; 

		IML_DOUBLE get_indexed_value( void );

}; // end class iml_rect_rofn


//=====================================================================
// THIS FUNCTION GENERATES THE VALUE
//  					Magnitude * sinc((X-D)*Wx)
//   ROTATED BY Phi DEGREES
//
//     WHERE		sinc(A*X) = sin(A * pi * x)/(A * pi * x)
//	    THIS MEANS THAT THE FIRST ZEROS OF THE FUNCTION ARE AT x = +/- (1/A)
//=====================================================================
class iml_sinc_rofn : public iml_rofn {

	private:
		IML_DOUBLE sinc(IML_DOUBLE Z) {
			if(	(Z < __IML_FNGEN_ALMOST_ZERO__) ) return(1.0);
			return( iml_sin( IML_PI * Z) / (IML_PI * Z) );
			};

		IML_DOUBLE sinc(IML_DOUBLE A, IML_DOUBLE Z) {
			if(	((A * Z) < __IML_FNGEN_ALMOST_ZERO__) ) return(1.0);
			return( iml_sin( A * IML_PI * Z) / (A * IML_PI * Z) );
			};

	public:

		iml_sinc_rofn( IML_ULINT DimCnt = 2 ) : iml_rofn(DimCnt) { };
		iml_sinc_rofn( iml_sinc_rofn IML_REF S ) : iml_rofn(S.Domain.Dimensionality) { };

		// WIDTH HERE IS THE TWICE THE DISTANCE TO THE FIRST ZERO FROM THE PEAK
		void set_width(IML_ULINT Dim, IML_DOUBLE Width) {
			set_halfwidth(Dim, Width/2.0);
         }; 

		// THIS IS THE SAME AS SETTING THE FIRST ZERO OF THE FUNCTION
		void set_halfwidth(IML_ULINT Dim, IML_DOUBLE HWidth) {
			if( HWidth < __IML_FNGEN_ALMOST_ZERO__ ) {
				iml_error("iml_sinc_rofn", "Halfwidth is too small.  Defaulting to 1.0.");
				HWidth = 1.0;
				} // end if
			set_scale(Dim, 1.0/HWidth);
         }; 

		~iml_sinc_rofn() {};
                                          
		IML_DOUBLE get_indexed_value( void );

}; // end class iml_sinc_rofn


#endif // __IML_ROFN_H__
