#ifndef __IML_SEQ_H__
#define __IML_SEQ_H__
/*----------------------------------------------------------------------
	Legal Stuff (tm)

	Copyright(c)1996-98 R.E. Fredericksen <eric@troland.hipl.uci.edu>
		U.C. Irvine, Irvine, CA (USA)
	
	The accompanying software is free; you may use it in your 
	research  and give it away to others, with the following 
	restrictions. Any copy  you give away must include this 
	Legal Stuff(tm) section, unmodified.  Any file that you have 
	changed must include a note at the top of the file giving your 
	name, the date, and a description of the changes. This software
	may not be sold, whether in source or compiled form, without 
	my permission. 	
	
	THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR
	IMPLIED WARRANTY.  IN PARTICULAR, NEITHER THE AUTHOR NOR THE 
	UNIVERSITY OF CALIFORNIA MAKE ANY REPRESENTATION OR WARRANTY OF 
	ANY KIND CONCERNING THE MERCHANTABILITY OF THIS SOFTWARE OR ITS
	FITNESS FOR ANY PARTICULAR PURPOSE.
	
----------------------------------------------------------------------*/

#include "iml_type.h"
#include "iml_rand.h"
#include "iml_math.h"

// COULD MAKE THIS A CLASS MEMBER, BUT MANY NON-C++ PEOPLE WOULD BE CONFUSED
typedef enum {LOGARITHMIC, LINEAR} iml_sequence_type;

//=====================================================================
//=====================================================================
/** This template class is a general purpose scalar-sequence manager.
It was written, really, because I wanted to make lists of unsigned
integers to be used for specifying frame sequences of movie loops.
This idea was stolen . . ahem . . borrowed from D. Brainard's
MatLab interface to D.G. Pelli's VideoToolbox library.

<p>  After creating the class I have predefined a sequence manager
for each scalar type:
<pre>

	// An 8-bit signed integer sequence manager.
	typedef iml_sequence<IML_BYTE>	iml_byte_sequence;
	
	// An 8-bit unsigned integer sequence manager.
	typedef iml_sequence<IML_UBYTE>	iml_ubyte_sequence;
	
	// A 16-bit signed integer sequence manager.
	typedef iml_sequence<IML_SINT>	iml_sint_sequence;
	
	// A 16-bit unsigned integer sequence manager.
	// This is the one we use to create the alias for
	// our image frame index manager.
	typedef iml_sequence<IML_USINT>	iml_usint_sequence;
	
	// A 32-bit signed integer sequence manager.
	typedef iml_sequence<IML_LINT>	iml_lint_sequence;
	
	// A 32-bit unsigned integer sequence manager.
	typedef iml_sequence<IML_ULINT>	iml_ulint_sequence;
	
	// A 32-bit floating point sequence manager. 
	typedef iml_sequence<IML_FLOAT>	iml_float_sequence;
	
	// A 64-bit floating point sequence manager.
	typedef iml_sequence<IML_DOUBLE>	iml_double_sequence;

</pre>

I then use
<pre>

	typedef iml_usint_sequence		iml_imageframe_sequence;

</pre>
as the type for an image frame sequence.
*/
template <class Type> class iml_sequence {

	protected:
		// DATA STUFF
		IML_USINT Length;
		Type IML_PTR Sequence;
		
		// A PRIVATE RANDOM NUMBER GENERATOR USED FOR SHUFFLING SEQUENCE ENTRIES
		uniform_rv Shuffler;

	public:
	

		/** @name A: The following are the constructors and destructors. </p>*/
		//@{
		/** A complete sequence constructor.
		The default is a linear sequence, although you may also specify
		a logarithmic (that is, multiplicative) sequence.
		*/
		iml_sequence(Type Start, Type End, 
			IML_USINT Len, iml_sequence_type SeqType= LINEAR);

		/** A simple sequence constructor.
		In this case you either construct the sequence yourself or
		use other methods to initialize the values.  You can use the
		square-bracket (that is, <em>operator[]</em>) to modify each
		value on your own.
		*/
		explicit iml_sequence(IML_USINT Len);
		
		explicit iml_sequence(iml_sequence<Type> & Seq);
		
		/* The only destructor. */
		~iml_sequence() { if(Sequence) delete [] Sequence; };

		//@}
		
		/** @name B: The following are methods used when constructing a sequence. </p>*/
		//@{

		/** Fills the array with a linear sequence beginning at <b>Start</b>
		and incremented by <b>Delta</b> along the way.
		*/
		void generate_linear_sequence(Type Start, Type Delta);
		/** Fills the array with a logarithmic (multiplicative) sequence 
		beginning at <b>Start</b> and multiplied by <b>Delta</b> along the way.
		*/
		void generate_logarithmic_sequence(Type Start, IML_DOUBLE Delta);
		
		/** Fills the array with a sinusoidally varying index sequence.
		If these indices are frame numbers then you can get an approximation
		of sinsoidally varying motion direction/speed. :-)  The <em>PhaseDeg</em>
		component might be unnecessary because you can also rotate the sequence,
		but it makes the specification a bit more complete.
		*/
		void sin(IML_DOUBLE Mean, IML_DOUBLE Amplitude, 
								IML_DOUBLE Period, IML_DOUBLE PhaseDeg);

		/** This method copies one sequence into another. If the destination sequence
		is shorter than the source, only the corresponding segment of the longer
		sequence is copied.  If the destination sequence is longer than the
		source (this is the handy part) then the shorter source sequence is
		replicated (repeated) in the longer destination sequence.
		*/
		void operator=(iml_sequence<Type> & Seq);
		
		/** This method provides direct access to a sequence entry.  Array
		bounds checking <b>is not done</b>, by default.  Use the <em>operator()</em>
		method to get array bounds checking.
		*/
		Type IML_REF operator[](IML_USINT I) { return(Sequence[I]); };

		/** This method provides direct access to a sequence entry.  Array
		bounds checking <b>is done</b>, by default.  Use the <em>operator()</em>
		method to get array bounds checking.
		*/
		Type IML_REF operator()(IML_USINT I) {
			return( Sequence[ ((I<Length) ? I : (Length-1))  ] );
			};
		
		/** This method returns the length of the array (sequence).  This is
		provides so we can loop and fill it ourselves, but not have to keep
		track of which sequences are which length.
		*/
		IML_USINT length(void) { return(Length); }

		//@}
		
		/** @name C: The following are methods for manipulating a previusly 
		constructed equence. 
		*/
		//@{

		// =========== WAYS TO MANIPULATE THE SEQUENCE ================
		/** Reverse the sequence by swapping the pairs (0,N-1), (1,N-2) etc.
		*/
		void reverse(void);

		/** Rotate the sequence by the number of specified places.
		Positive values produce rotation to the left and negative values 
		produce rotation to the right.  Thus, for a sequence
		<pre>
		
			0 1 2 3 4 5
		
		</pre>
		
		a call <em>rotate(2)</em> produces the new sequence
		<pre>
		
			2 3 4 5 0 1
		
		</pre>
		
		*/
		void rotate(IML_SINT Amount);
		
		/** Shuffles the sequence of values.  The algorithm is just
		<pre>
			for length() times do
				create a pair of random integers between 0 and length()-1
				swap those entries in the sequence
		</pre>
		
		The default is to shuffle just once, but it is possible to to shuffle 
		the sequence an arbitrary number of times.  A zero parameter value leaves 
		the sequence unchanged.
		*/
		void shuffle(IML_USINT Times = 1);
		
		//@}

		// =========== WAYS TO ACCESS THE SEQUENCE =====================
		// ACCESS MEMBER FUNCTIONS;  THE USUAL APPLIES: [] IS FOR
		// FAST, UNSAFE ACCESS, AND () IS FOR BOUNDS-CHECKED ACCESS.
		
}; // end class sequence


//=====================================================================
//
//=====================================================================
// GENERIC PRE-DEFINED TEMPLATES
/* An 8-bit signed integer sequence manager.*/
typedef iml_sequence<IML_BYTE>	iml_byte_sequence;
/* An 8-bit unsigned integer sequence manager.*/
typedef iml_sequence<IML_UBYTE>	iml_ubyte_sequence;
/* A 16-bit signed integer sequence manager.*/
typedef iml_sequence<IML_SINT>	iml_sint_sequence;
/* A 16-bit unsigned integer sequence manager.
This is the one we use to create the alias for our image frame index manager.
*/
typedef iml_sequence<IML_USINT>	iml_usint_sequence;
/* A 32-bit signed integer sequence manager.*/
typedef iml_sequence<IML_LINT>	iml_lint_sequence;
/* A 32-bit unsigned integer sequence manager.*/
typedef iml_sequence<IML_ULINT>	iml_ulint_sequence;
/* A 32-bit floating point sequence manager. */
typedef iml_sequence<IML_FLOAT>	iml_float_sequence;
/* A 64-bit floating point sequence manager.*/
typedef iml_sequence<IML_DOUBLE>	iml_double_sequence;

/** This is an image frame index manager created using the 
<em>iml_sequence</em> template class and the IML_USINT scalar
type. The assumption is that we won't want an index sequence longer than
(2**16)-1 some time in the future. ('**' denotes exponentiation)
*/
typedef iml_usint_sequence 		iml_imageframe_sequence;


// THESE ARE REQUIRED HERE BECAUSE CODE-WARRIOR DOES TEMPLATE
// CODE GENERATION DIFFERENTLY THAN BORLAND DOES. . . WHAT A PAIN.
#ifdef __BORLAND_CPP__
#include "iml_seq.cpp"
#endif


#endif //__IML_SEQ_H__


