#ifndef __ROTMAT3d_H__
#define __ROTMAT3d_H__

#include "vector_3d.h"

// MAKE SURE YOU INCLUDE AND/OR DEFINE YOUR POINT CLASS BEFORE
// THIS HEADER FILE.!!!


//===========================================================
//
//        Here is a 3x3 matrix that knows how to post-multiply
//        itself by a 3x1 vector_3d vector.  The matrix knows how to
//        produce an identity matrix, and a rotation matrix.
//        The rotation matrix requires a vector describing the
//        axis through the origin, and the rotation angle in 
//        radians.  Remember:  RADIANS. :-)
//
//===========================================================
class matrix_3d {
	private:
		float M[3][3];
		
	public:
	
		// The single contructor.
		matrix_3d(void) { identity(); }; 
		
		// Make an Identity matrix, of course.
		void identity(void) {
			for(long X=0; X<3; X++)
				for(long Y=0; Y<3; Y++)
					M[X][Y] = (X==Y) ? 1 : 0;
			};

		// Peforms the operation Point = Point*Matrix.
		//  [Px Py Pz] * [  M00 M01 M02  ]
		//               [  M10 M11 M12  ]
		//               [  M20 M21 M22 ]
		void post_multiply(vector_3d & P) {
			float X = P.x(), Y=P.y(), Z=P.z();
			P.x( X*M[0][0] + Y*M[1][0] + Z*M[2][0] );
			P.y( X*M[0][1] + Y*M[1][1] + Z*M[2][1] );
			P.z( X*M[0][2] + Y*M[1][2] + Z*M[2][2] );
			};
			
		// Fills in the appropriate rotation matrix for
		// rotation about an arbitrary Axis by AngleInRadians.
		void set_rotation(vector_3d Axis, float AngleInRadians) {
			Axis.normalize();
			float X = Axis.x(), Y = Axis.y(), Z = Axis.z();
			float Cos = cos(AngleInRadians);
			float Sin = sin(AngleInRadians);
			M[0][0] = X*X + (1-X*X) * Cos;
			M[1][1] = Y*Y + (1-Y*Y) * Cos;
			M[2][2] = Z*Z + (1-Z*Z) * Cos;
			
			M[0][1] = X*Y * (1-Cos) + Z*Sin;
			M[1][0] = X*Y * (1-Cos) - Z*Sin;
			
			M[0][2] = X*Z * (1-Cos) - Y*Sin;
			M[2][0] = X*Z * (1-Cos) + Y*Sin;
			
			M[1][2] = Y*Z * (1-Cos) + X*Sin;
			M[2][1] = Y*Z * (1-Cos) - X*Sin;
			
			};

		// Fills in the appropriate rotation matrix for
		// rotation about an axis defined by two vectors,
		// Pa and Pb.  The two vector_3ds define both an axis,
		// and a rotation angle.  The two vector_3ds are converted
		// to unit vectors, and hence vector_3ds lying on
		// the surface of the unit sphere.  This allows us to
		// avoid the case where the two vectors are co-linear,
		// and where there is no unique axis of rotation.
		void set_rotation( vector_3d Pa, vector_3d Pb) {
		
			Pa.normalize();
			Pb.normalize();
			
			// case of no rotation; two vector_3ds are the same
			if(Pa.x() == Pb.x() && Pa.y() == Pb.y() && Pa.x() == Pb.x() ) {
				identity();
				return;
				} // end if
				
			vector_3d Axis = Pa.cross( Pb );
			float Rotation = acos( Pa.dot(Pb) );
			set_rotation(Axis, Rotation);
			
			};


	}; // end class matrix_3d


#endif // __ROTMAT3d_H__
