#ifndef __SPHEREMAP_H__
#define __SPHEREMAP_H__

#include "vector_3d.h"

//===========================================================
//
//        This class provides a mapping from a unit circle
//        to the surface of the unit sphere.  Given the
//        circle center and the circle diameter (the size
//        of the bounding box) the function map_to_unit_vector()
//        takes (X,Y) positions in the box to the corresponding
//        vector_3d on the unit sphere.
//
//===========================================================
class sphere_map {
	private:
		float HalfXYRadius, XCenter, YCenter;
	
	public:
	
		// The constructor requries the center and the diameter.
		sphere_map( float XC, float YC, float R ) :
			Radius(R), XCenter(XC), YCenter(YC)
			{};
		
		// This is the circle->sphere mapper.
		vector_3d map_to_unit_vector(float X, float Y) {
			
			// Special case: the vector_3ds are co-linear.
			if( (X == XCenter) && (Y == YCenter)) return( vector_3d(0,0,1.0) );
			
			// Convert (X,Y) to circle-centered coords.
			X -= XCenter;	Y -= YCenter;
			
			// Get the X-Y plane angle defined by (X,Y).
			float Theta = atan2(X,Y);
			
			// Calculate distance from circle center.
			float XYRadius = sqrt(X*X + Y*Y);
			
			// Map the (X,Y) radius distance onto (90,  -90) deg range.
			// Inner half-radius maps to front face of sphere,
			// while the outer half-radius maps to the back face.
			float Phi = 90.0 - 180.0*XYRadius/Radius;
			
			// Map all of outside of circle to the equator.
			if(Phi <= -90.0) Phi = 0.0;
			
			// Trig functions expect radians, not degrees.
			Phi *= 3.14159265359/180.0;
			
			// Calculate the (X,Y,Z) vector.
			float Px = cos(Theta)*cos(Phi);
			float Py = sin(Theta)*cos(Phi);
			float Pz = sin(Phi);
			
			// Return the answer.
			return( vector_3d(Px,Py,Pz) ) ;
			};

}; // end class sphere_map

#endif // __SPHEREMAP_H__