//=====================================================================
// THESE ARE CLASSES INTENDED TO MAKE DESIGNING AND CONTROLLING
// EXPERIMENTS EASIER. :-)
//=====================================================================

#ifndef __IML_EXPS_H__
#define __IML_EXPS_H__

#include <stdio.h>
#include "iml_type.h"
#include "iml_math.h"
#include "iml_err.h"
//#include "values.h"
#include "iml_rand.h"


//=====================================================================
// GENERIC DOUBLY LINKED LIST ITEM
//=====================================================================
class dl_list_item {
	public:

		dl_list_item IML_PTR Next;
		dl_list_item IML_PTR Prev;

		// CONSTRUCTOR
		dl_list_item(	dl_list_item IML_PTR NextI = NULL,
							dl_list_item IML_PTR PrevI = NULL) {
			Next = NextI;
			Prev = PrevI;
			};

		void add(		dl_list_item IML_PTR NextI = NULL,
							dl_list_item IML_PTR PrevI = NULL) {
			Next = NextI;
         if(!Next) Next->Prev = this;
			Prev = PrevI;
			if(!Prev) Prev->Next = this;
         };

		void remove(void) {
			Next->Prev = Prev;
			Prev->Next = Next;
			};

		virtual IML_DOUBLE sort_value(void) {
			iml_error("dl_list_item", "Call to base class sort_value()!! Need to replace!!");
			return(0);
         };

}; // end class



//=====================================================================
// A MANAGER FOR GENERIC LISTS OF DOUBLY LINKED ITEMS
//=====================================================================
template <class DLLType> class dl_list_manager {
	private:
		DLLType Safety;
		dl_list_item IML_PTR Current;

	protected:
		DLLType Head;
		DLLType Tail;

	public:
   	IML_USINT DatumCount;

		dl_list_manager(void) {
      	// BETTER TO HAVE AN INFINITE LOOP THAN A CRASH...
			Head.Prev = &Tail;
			Head.Next = &Tail; // DO NOT STACK ASSIGNS; COMPILER LOSES HUGE PTR NORMALIZATION!!
			Tail.Prev = &Head; // i.e. DO NOT DO Head.Prev = Head.Next = &Tail;
			Tail.Next = &Head;

			// USE THESE VALUES TO DO ORDERED INSERTION BETWEEN HEAD AND TAIL
			Head.ParameterValue = -DBL_MAX;
			Tail.ParameterValue =  DBL_MAX;

			DatumCount = 0;
         Current = NULL;
			};


		~dl_list_manager(void) { clear_data(); };


		// THE FOLLOWING ARE THREE FUNCTIONS TO ACCESS THE ACCUMULATED DATA:
		// THE FIRST TWO ARE FOR SEQUENTIAL ACCESS:  'first' AND 'next';  END
		// OF LIST IS SIGNALLED BY A NULL PTR.
		// THE SECOND IS A SOMEWHAT INEFFICIENT RANDOM ACCESS OPERATOR THAT HAS A
		// CONVENIENT NOTATION.
		DLLType IML_PTR first(void) {
			if(DatumCount) Current = Head.Next;
			return((DLLType IML_PTR) Current);
			};

		DLLType IML_PTR next(void) {
      	// IF NOT ENDED ALREADY
			if(Current) {
				Current = Current->Next;
				// IF WE ARE AT THE END OF THE LIST...(Head.Prev == &Tail)
				if(Current == Head.Prev) Current = NULL;
				}; // end if
			return((DLLType IML_PTR) Current);
			};

      
		DLLType IML_REF operator[](IML_LINT Index);


		void clear_data(void);


		DLLType IML_PTR find_or_add_item(IML_DOUBLE ParmValue);

}; // end class



//=====================================================================
// THIS IS A BASE CLASS FOR WRITING USER RESPONSE-GETTER FUNCTIONS/OBJECTS
//=====================================================================
class user_input {
	public:
		virtual IML_BYTE abort_requested(void) {
			iml_error("user_TF_input",
				"Call to abort_requested() in base class.  Should be replaced in derived class!!!");
			return(TRUE);
		};

//		virtual IML_BYTE repeat_requested(void) {
//			iml_error("user_TF_input",
//				"Call to repeat_requested() in base class.  Should be replaced in derived class!!!");
//			return(FALSE);
//		};


		virtual IML_USINT response(IML_SINT InFlags = 0) {
			iml_error("user_TF_input",
				"Call to response() in base class.  Should be replaced in derived class!!!");
			return(TRUE);
			};
}; // end class




//=====================================================================
// STIMULI TO BE USED IN AN EXPERIMENT (SEE BELOW) SHOULD DERIVE FROM THIS VIRTUAL BASE
// CLASS;  IT PROVIDES A GENERIC WAY TO ADJUST THE STIMULUS IN AN EXPERIMENT WITHOUT
// HAVING TO KNOW WHICH PARAMETER IS BEING ADJUSTED.
//=====================================================================
class adjustable_stimulus {
	private:
	adjustable_parameter Safety;

	public:

		virtual adjustable_parameter IML_REF threshold_variable(void) {
			iml_error("adjustable_stimulus",
					"Call to threshold_variable() in base class!! Replace me!!");
			return(Safety);
			};

		virtual IML_USINT present(IML_SINT Flags = 0) {
			iml_error("adjustable_stimulus",
					"Call to present() in base class!! Replace me!!");
         return(0);
			};

}; // end class



//=====================================================================
// STAIRCASE EXPERIMENT BASE CLASS; DERIVED CLASSES CAN REPLACE THE VIRTUAL
// FUNCTIONS
//=====================================================================
class staircase_experiment {


	public:
		virtual IML_BYTE abort_requested(void) {
			iml_error("staircase_experiment",
				"Call to abort_requested() in base class.  Should be replaced in derived class!!!");
			return(TRUE);
			};

		// DEFAULT STAIRCASE DOES NOT REALLY KNOW, BUT QUEST DOES.
		virtual IML_SINT estimated_steps(void) { return(50); };

		// THIS IS TO ALLOW DERIVED CLASSES TO BE OPERATED :-)
		virtual void make_a_step(IML_SINT Flags = 0) {
			iml_error("staircase_experiment",
								"Call to make_a_step should be replaced in derived class!!");
			};
		virtual IML_BYTE staircase_complete(void) {
			iml_error("staircase_experiment",
								"Call to staircase_complete should be replaced in derived class!!");
			return(TRUE);
			};

}; // end class

//=====================================================================
// BLOCK STAIRCASE EXPERIMENT CLASS
//=====================================================================
class block_staircase_experiment {
	private:
		staircase_experiment Safety;
		staircase_experiment IML_PTR SafetyPtr;
		IML_SINT  Indices[256];
		IML_FLOAT Distribution[256];

	protected:
		staircase_experiment IML_PTR *Block;
		IML_BYTE Count;
		IML_UBYTE random_index(void) {
			return((IML_UBYTE) (UniformRV.sample() * Count) );
         };

	public:
		uniform_rv UniformRV;

		block_staircase_experiment(IML_UBYTE Size) {
			SafetyPtr = &Safety;
			UniformRV.set_min_max(0.0,0.999999999999999);
			UniformRV.seed(-1234);
			Count = Size;
			Block = new(staircase_experiment IML_PTR[Size]);
			if( !Block ) {
				iml_error("block_staircase_experiment", "Out of memory! Could not allocate array!!");
				} // end if
			}; // constructor

		~block_staircase_experiment(void) {
			if(Block) delete [] Block;
			}; // destructor

		IML_BYTE block_complete(void);

		// ONE PRESENTATION, ONE STIMULUS, CHOSEN RANDOMLY
		void make_a_random_step(void);

		// ONE TRIAL FOR EACH STIMULUS, RANDOMIZED ORDER
		void make_a_block_step(void);

		// A BLOCK OF Size PRESENTATIONS, STIMULI CHOSEN AT RANDOM
		void make_a_random_block_step(IML_UBYTE Size);

		staircase_experiment IML_PTR & operator[](IML_UBYTE Index) {
			if( !Block ) {
				iml_error("block_staircase_experiment", "Access after memory allocation failure!");
				return( SafetyPtr );
				} // end if
			if(Index >= Count) {
				iml_error("block_staircase_experiment","Index out of range!!!!!");
				return(SafetyPtr);
				} // end if
			return(Block[Index]);
			};
}; // end class


//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
//
// THE ITEMS BELOW ARE PLACED THERE SO THAT THE CODE WILL BE GENERATED
// CORRECTLY FOR THE TEMPLATE INSTANTIATIONS FOR THE dl_list_manager
//
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$


//=====================================================================
// THIS ONE IS A HOLDER FOR "FREQUENCY OF SEEING" (FOS) FOR A GIVEN
// THRESHOLD PARAMETER VALUE;  THE IDEA IS TO KEEP AN ORDERED LIST OF
// THESE THINGS, INCREMENTING VALUES AS YOU GO.
//=====================================================================
class fos_item : public dl_list_item {

	public:
		IML_DOUBLE ParameterValue;
		IML_LINT	  Seen;
		IML_LINT   NotSeen;

		// CONSTRUCTOR
		fos_item( IML_DOUBLE Value = 0.0,
						dl_list_item IML_PTR NextFOS = NULL,
						dl_list_item IML_PTR PrevFOS = NULL) : dl_list_item(NextFOS,PrevFOS) {
			ParameterValue = Value;
			Seen = NotSeen = 0;
			};

		// OBLIGATORY SORTING VALUE ROUTINE
		IML_DOUBLE sort_value(void) { return(ParameterValue); };

}; // end class


//============================================================================
// INSTANTIATIONS OF SPECIFIC ITEM LIST MANAGERS
//============================================================================
class fos_data_manager : public dl_list_manager<fos_item>  {

	public:

		void add_datum(IML_DOUBLE ParmValue, IML_BYTE SeenTrue) {

			fos_item IML_PTR Here = find_or_add_item(ParmValue);

			// INCREMENT THE FOS;  ONLY ONE OF THESE WILL INCREMENT
			Here->Seen 	  += (SeenTrue ? 1 : 0);
			Here->NotSeen += (SeenTrue ? 0 : 1);

			};

}; // end  class



//===================================================================================
// HOLDER OF INFORMATION FOR WILLIAM MCILHAGGA'S QUEST ALGORITHM
//===================================================================================
class wm_quest_item : public fos_item {

	public:
		IML_DOUBLE PriorL;
		IML_DOUBLE SumLogL;
		IML_DOUBLE Temp;

		wm_quest_item(IML_DOUBLE Value = 0.0,
						dl_list_item IML_PTR NextFOS = NULL,
						dl_list_item IML_PTR PrevFOS = NULL) :
						fos_item(Value,NextFOS,PrevFOS) {

			PriorL = SumLogL = Temp = 0.0;
			};

}; // end class



//============================================================================
// INSTANTIATIONS OF WILLIAM'S QUEST ITEM LIST MANAGERS
//============================================================================
class wm_quest_data_manager : public dl_list_manager<wm_quest_item>  {

	public:

		void add_datum(IML_DOUBLE ParmValue, IML_BYTE SeenTrue) {

			wm_quest_item IML_PTR Here = find_or_add_item(ParmValue);

			// INCREMENT THE FOS;  ONLY ONE OF THESE WILL INCREMENT
			Here->Seen 	  += (SeenTrue ? 1 : 0);
			Here->NotSeen += (SeenTrue ? 0 : 1);

			};

}; // end class



#endif  // __IML_EXPS_H__


