/*
 * Copyright (c) 1997 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/* fftw.h -- system-wide definitions */
/* $Id: fftw.h,v 1.28 1997/05/02 08:22:00 fftw Exp $ */

#include <stdlib.h>

/* our real numbers */
typedef double FFTW_REAL;

/*********************************************
 * Complex numbers and operations 
 *********************************************/
typedef struct {
     FFTW_REAL re, im;
} FFTW_COMPLEX;

#define c_re(c)  ((c).re)
#define c_im(c)  ((c).im)

typedef enum {
     FFTW_FORWARD = -1, FFTW_BACKWARD = 1
} fftw_direction;

#ifndef FFTW_1_0_COMPATIBILITY
#define FFTW_1_0_COMPATIBILITY 1
#endif

#if FFTW_1_0_COMPATIBILITY
/* backward compatibility with FFTW-1.0 */
#define REAL FFTW_REAL
#define COMPLEX FFTW_COMPLEX
#endif

/*********************************************
 *              Codelets
 *********************************************/
/*
 * There are two kinds of codelets:
 *
 * NO_TWIDDLE    computes the FFT of a certain size, operating
 *               out-of-place (i.e., take an input and produce a
 *               separate output)
 *
 * TWIDDLE       like no_twiddle, but operating in place.  Moreover,
 *               multiplies the input by twiddle factors.
 */

typedef void (notw_codelet) (const FFTW_COMPLEX *, FFTW_COMPLEX *, int, int);
typedef void (twiddle_codelet) (FFTW_COMPLEX *, const FFTW_COMPLEX *, int,
				int, int);
typedef void (generic_codelet) (FFTW_COMPLEX *, const FFTW_COMPLEX *, int,
				int, int, int);

/*********************************************
 *     Configurations
 *********************************************/
/*
 * A configuration is a database of all known codelets
 */

typedef struct {
     int size;			/* size of the problem */
     notw_codelet *codelet;	/* 
				 * pointer to the codelet that solves the 
				 * problem 
				 */
} config_notw;

extern config_notw fftw_config_notw[];
extern config_notw fftwi_config_notw[];

typedef struct {
     int size;			/* size of the problem */
     twiddle_codelet *codelet;
} config_twiddle;

extern config_twiddle fftw_config_twiddle[];
extern config_twiddle fftwi_config_twiddle[];

extern generic_codelet fftw_twiddle_generic;
extern generic_codelet fftwi_twiddle_generic;
extern char *fftw_version;

/*****************************
 *        Plans
 *****************************/
/*
 * A plan is a sequence of reductions to compute a FFT of
 * a given size.  At each step, the FFT algorithm can:
 *
 * 1) apply a notw codelet, or
 * 2) recurse and apply a twiddle codelet, or
 * 3) apply the generic codelet.
 */

enum fftw_node_type {
     FFTW_NOTW, FFTW_TWIDDLE, FFTW_GENERIC
};

/* structure that contains twiddle factors */
typedef struct fftw_twiddle_struct {
     int n;
     int r;
     int m;
     FFTW_COMPLEX *twarray;
     struct fftw_twiddle_struct *next;
     int refcnt;
} fftw_twiddle;

/* structure that holds all the data needed for a given step */
typedef struct fftw_plan_node_struct {
     enum fftw_node_type type;

     union {
	  /* nodes of type FFTW_NOTW */
	  struct {
	       int size;
	       notw_codelet *codelet;
	  } notw;

	  /* nodes of type FFTW_TWIDDLE */
	  struct {
	       int size;
	       twiddle_codelet *codelet;
	       fftw_twiddle *tw;
	       struct fftw_plan_node_struct *recurse;
	  } twiddle;

	  /* nodes of type FFTW_GENERIC */
	  struct {
	       int size;
	       generic_codelet *codelet;
	       fftw_twiddle *tw;
	       struct fftw_plan_node_struct *recurse;
	  } generic;

     } nodeu;

     int refcnt;
} fftw_plan_node;

struct fftw_plan_struct {
     int n;
     fftw_direction dir;
     fftw_plan_node *root;

     double cost;
     int flags;

     struct fftw_plan_struct *next;
     int refcnt;
};

/* a plan is just an array of instructions */
typedef struct fftw_plan_struct *fftw_plan;

/* flags for the planner */
#define  FFTW_ESTIMATE (0)
#define  FFTW_MEASURE  (1)

#define FFTW_IN_PLACE (8)

extern fftw_plan fftw_create_plan(int n, fftw_direction dir, int flags);
extern fftw_twiddle *fftw_create_twiddle(int n, int r, int m);
extern void fftw_destroy_twiddle(fftw_twiddle * tw);
extern void fftw_print_plan(fftw_plan plan);
extern void fftw_destroy_plan(fftw_plan plan);
extern void fftw_naive(int n, FFTW_COMPLEX *in, FFTW_COMPLEX *out);
extern void fftwi_naive(int n, FFTW_COMPLEX *in, FFTW_COMPLEX *out);
void fftw(fftw_plan plan, int howmany, FFTW_COMPLEX *in, int istride,
	  int idist, FFTW_COMPLEX *out, int ostride, int odist);
extern double fftw_measure_runtime(fftw_plan plan);
extern void fftw_die(char *s);
extern void *fftw_malloc(size_t n);
extern void fftw_free(void *p);
extern void fftw_strided_copy(int, FFTW_COMPLEX *, int, FFTW_COMPLEX *);
extern void fftw_executor_simple(int, const FFTW_COMPLEX *, FFTW_COMPLEX *,
				 fftw_plan_node *, int, int);
extern void *(*fftw_malloc_hook) (size_t n);
extern void (*fftw_free_hook) (void *p);

/*****************************
 *    N-dimensional code
 *****************************/
typedef struct {
     int is_in_place;		/* 1 if for in-place FFT's, 0 otherwise */
     int rank;			/* 
				 * the rank (number of dimensions) of the
				 * array to be FFT'ed
				 */
     int *n;			/* 
				 * the dimensions of the array to the
				 * FFT'ed 
				 */
     int *n_before;		/* 
				 * n_before[i] = product of n[j] for j < i 
				 */
     int *n_after;		/* n_after[i] = product of n[j] for j > i */
     fftw_plan *plans;		/* fftw plans for each dimension */
     FFTW_COMPLEX *work;	/* 
				 * work array for FFT when doing
				 * "in-place" FFT 
				 */
} fftwnd_aux_data;

typedef fftwnd_aux_data *fftwnd_plan;

/* Initializing the FFTWND Auxiliary Data */
fftwnd_plan fftw2d_create_plan(int nx, int ny, fftw_direction dir, int flags);
fftwnd_plan fftw3d_create_plan(int nx, int ny, int nz,
			       fftw_direction dir, int flags);
fftwnd_plan fftwnd_create_plan(int rank, const int *n, fftw_direction dir,
			       int flags);

/* Freeing the FFTWND Auxiliary Data */
void fftwnd_destroy_plan(fftwnd_plan plan);

/* Computing the N-Dimensional FFT */
void fftwnd(fftwnd_plan plan, int howmany,
	    FFTW_COMPLEX *in, int istride, int idist,
	    FFTW_COMPLEX *out, int ostride, int odist);

/*************************************
 *  Timers, etc
 *************************************/
/* 
 * Here you can use all the nice timers available
 * in your machine
 */

#ifdef SOLARIS

/* we use the nanosecond virtual timer */
#include <sys/time.h>

typedef hrtime_t fftw_time;

#define fftw_get_time() gethrtime()
#define fftw_time_to_sec(t) ((double) (t) / 1.0e9)

/* the clock should be precise enough for this number of iterations */
#define FFTW_PLAN_ITER 10000

#endif

#if defined(MAC) || defined(macintosh)
double get_Mac_microseconds();

typedef double fftw_time;

#define fftw_get_time() get_Mac_microseconds()
#define fftw_time_to_sec(t) (t)

/* the clock should be precise enough for this number of iterations */
#define FFTW_PLAN_ITER 100000
#endif

#ifdef __WIN32__

#include <time.h>

typedef unsigned long fftw_time;
extern unsigned long GetPerfTime(void);
extern double GetPerfSec(double ticks);

#define fftw_get_time() GetPerfTime()
#define fftw_time_to_sec(t) GetPerfSec(t)

#ifndef CLOCKS_PER_SEC
#error Please define CLOCKS_PER_SEC
#endif

/* A conservative number if high-performance counter is indeed installed */
#define FFTW_PLAN_ITER 1000000
#endif				/* __WIN32__ */

/*
 * last resort: good old Unix clock()
 */
#ifndef fftw_get_time
#include <time.h>

typedef clock_t fftw_time;

#ifndef CLOCKS_PER_SEC
#ifdef sun
/* stupid sunos4 prototypes */
#define CLOCKS_PER_SEC 1000000
extern long clock(void);
#else
#error Please define CLOCKS_PER_SEC
#endif
#endif

#define fftw_get_time() clock()
#define fftw_time_to_sec(t) ((double) (t) / CLOCKS_PER_SEC)

#define FFTW_PLAN_ITER 1000000
#endif
