/*
 * Copyright (c) 1997 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *
 */

#include <fftw.h>

/************************* Thread Glue *************************/

/* Adding support for a new shared memory thread API should be easy.  You
   simply do the following things (look at the POSIX and Solaris
   threads code for examples):

   * Invent a symbol of the form FFTW_USING_FOO_THREADS to denote
     the use of your thread API, and add an 
              #elif defined(FFTW_USING_FOO_THREADS)
      before the #else clause below.  This is where you will put
      your thread definitions.  In this #elif, insert the following:

      -- #include any header files needed to use the thread API.

      -- Typedef fftw_thread_function to be a function pointer
         of the type used as a argument to fftw_thread_spawn
	 (i.e. the entry function for a thread).

      -- Define fftw_thread_id, via a typedef, to be the type
         that is used for thread identifiers.

      -- #define fftw_thread_spawn(tid_ptr, proc, data) to
         call whatever function to spawn a new thread.  The
         new thread should call proc(data) as its starting point,
         and tid_ptr is a pointer to a fftw_thread_id that
         is set to an identifier for the thread.  You can also
         define this as a subroutine (put it in fftw_threads.c)
	 if it is too complicated for a macro.  The prototype should
	 be:

	 void fftw_thread_spawn(fftw_thread_id *tid_ptr,
	                        fftw_thread_function proc,
				void *data);

      -- #define fftw_thread_wait(tid) to block until the thread
         whose identifier is tid has terminated.  You can also
         define this as a subroutine (put it in fftw_threads.c) if
	 it is too complicated for a macro.  The prototype should be:
	
	 void fftw_thread_wait(fftw_thread_id tid);

   * If you need to perform any initialization before using threads,
     put your initialization code in the fftw_threads_init() function
     in fftw_threads.c, bracketed by the appropriate #ifdef of course.

   * Also, of course, you should modify the Makefile so that
     FFTW_THREADS = -DFFTW_USING_FOO_THREADS (and adds anything
     to the include path that is necessary) and set FFTW_THREADLIB
     to include any libraries necessary for using your threads.

   * Finally, if you do implement support for a new threads API, be
     sure to let us know at fftw@theory.lcs.mit.edu so that we can
     distribute your code to others!

*/
      
#if defined(FFTW_USING_SOLARIS_THREADS)

/* Solaris threads glue.  Tested. */

/* link with -lthread */

#include <thread.h>

/* Thread entry point: */
typedef void * (*fftw_thread_function) (void *);

typedef thread_t fftw_thread_id;

#define fftw_thread_spawn(tid_ptr, proc, data) \
     thr_create(0,0,proc,data,THR_BOUND,tid_ptr)

#define fftw_thread_wait(tid) thr_join(tid,0,0)

#elif defined(FFTW_USING_BEOS_THREADS)

/* BeOS threads glue.  NOT YET TESTED!  
   (I just read the BeBook documentation...hopefully, this
   will work, but I provide no guarantees.  Consider it
   as a starting point for your own implementation.) */

#include <OS.h>

/* Thread entry point: */
typedef thread_entry fftw_thread_function;

typedef thread_id fftw_thread_id;

#define fftw_thread_spawn(tid_ptr, proc, data) { \
     *(tid_ptr) = spawn_thread(proc,"FFTW",B_NORMAL_PRIORITY,data); \
     resume_thread(*(tid_ptr)); \
}

/* wait_for_thread requires that we pass a valid pointer as the
   second argument, even if we're not interested in the result. */
#define fftw_thread_wait(tid) {long exit_val;wait_for_thread(tid, &exit_val);}

#else /* use the default, POSIX threads: */

/* POSIX threads glue.  Tested. */

#ifndef FFTW_USING_POSIX_THREADS
#define FFTW_USING_POSIX_THREADS
#endif

/* link with -lpthread */

#include <pthread.h>

/* Thread entry point: */
typedef void * (*fftw_thread_function) (void *);

extern pthread_attr_t fftw_pthread_attributes;

typedef pthread_t fftw_thread_id;

#define fftw_thread_spawn(tid_ptr, proc, data) \
     pthread_create(tid_ptr,&fftw_pthread_attributes,proc,data)

#define fftw_thread_wait(tid) pthread_join(tid,0)

#endif

extern int fftw_threads_init(void);

/************************ Parallel Loops ***********************/

#include <alloca.h>

/* The following macro is invoked to distribute a loop from
   0 to loopmax over nthreads threads.  Each thread is started
   by calling proc.  data is a pointer to an array of nthreads
   data structures, one of which is passed to each thread.  Two
   of the fields of the data structure should be named "min"
   and "max".  These tell proc to do the portion of the loop
   from min to max - 1.

   N.B. This macro actually spawns nthreads-1 threads.  The
   last invocation of proc is done on the calling thread.

   Note also that this macro does not return until the loop
   is completed (i.e. until all of the threads have terminated). 

   Finally, note that after this macro is called, data points
   to the last element in the original data array. */

#define fftw_thread_spawn_loop(loopmax,nthreads,proc,data) { \
  int work_size = loopmax / nthreads; \
  fftw_thread_id *tid = alloca(sizeof(fftw_thread_id) * (--nthreads)); \
  int i; \
  for (i = 0; i < nthreads; ++i) { \
    data->max = (data->min = i * work_size) + work_size; \
    fftw_thread_spawn(&tid[i],(fftw_thread_function)proc,(void *)data); \
    data += 1; \
  } \
  data->min = i * work_size; \
  data->max = loopmax; \
  proc(data); \
  for (i = 0; i < nthreads; ++i) \
    fftw_thread_wait(tid[i]); \
}

/******************** User Interface *********************/

extern void fftw_threads(int nthreads,
		  fftw_plan plan, int howmany, FFTW_COMPLEX *in, int istride,
		  int idist, FFTW_COMPLEX *out, int ostride, int odist);
extern void fftwnd_threads(int nthreads,
			   fftwnd_plan plan, int howmany,
			   FFTW_COMPLEX *in, int istride, int idist,
			   FFTW_COMPLEX *out, int ostride, int odist);

